import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal, assert_check_totals

step = 1e-6
size = 3

class Mult(om.ExplicitComponent):

    def setup(self):

        self.add_input('x', np.ones(size))
        self.add_input('y', np.ones(size))

        self.add_output('z', np.ones(size))

        self.declare_partials(of='*', wrt='*')

    def compute(self, inputs, outputs):

        outputs['z'] = inputs['x'] * inputs['y']

    def compute_partials(self, inputs, partials):

        partials['z', 'x'] = inputs['y']
        partials['z', 'y'] = inputs['x']


class GeometryAndAero(om.Group):

    def setup(self):

        self.add_subsystem("comp1", Mult(), promotes_inputs=['y'])
        self.add_subsystem("comp2", Mult(), promotes_inputs=['y'])
        self.add_subsystem("comp3", Mult(), promotes_inputs=['y'])

        # self.connect('comp1.z', 'comp2.x')
        self.connect('comp2.z', 'comp3.x')

        if self.method == 'fd':
            self.approx_totals(step=step, step_calc="abs", method=self.method, form="forward")
        else:
            self.approx_totals(method=self.method)


class TestSemiTotals(unittest.TestCase):

    def test_semi_totals_fd(self):
        prob = om.Problem()

        sub = prob.model.add_subsystem('sub', GeometryAndAero(), promotes=['*'])
        sub.method = 'fd'

        prob.model.add_design_var("y")
        prob.model.add_objective("comp3.z", index=0)

        prob.setup(force_alloc_complex=True, check=False)
        prob.set_val("y", 5.0 * np.ones(size))

        prob.run_model()

        # Deriv should be 75. Analytic was wrong before the fix.
        data = prob.check_totals(method="fd", form="forward", step=step, step_calc="abs", out_stream=None)
        assert_check_totals(data, atol=1e-5, rtol=1e-6)

    def test_semi_totals_cs(self):
        prob = om.Problem()

        sub = prob.model.add_subsystem('sub', GeometryAndAero(), promotes=['*'])
        sub.method = 'cs'

        prob.model.add_design_var("y")
        prob.model.add_objective("comp3.z", index=0)

        prob.setup(force_alloc_complex=True, check=False)
        prob.set_val("y", 5.0 * np.ones(size))

        prob.run_model()

        # Deriv should be 75. Analytic was wrong before the fix.
        data = prob.check_totals(method="cs", out_stream=None)

        assert_check_totals(data, atol=1e-6, rtol=1e-6)

    def test_semi_totals_cs_indirect(self):
        prob = om.Problem()

        prob.model.add_subsystem('indeps', om.IndepVarComp('yy', np.ones(size)))
        prob.model.add_subsystem('comp', om.ExecComp('z=2*y', z=np.ones(size), y=np.ones(size)))
        sub = prob.model.add_subsystem('sub', GeometryAndAero(), promotes=['*'])
        sub.method = 'cs'

        prob.model.connect('indeps.yy', 'comp.y')
        prob.model.connect('comp.z', 'y')
        prob.model.add_design_var("indeps.yy")
        prob.model.add_objective("comp3.z", index=0)

        prob.setup(force_alloc_complex=True, check=False)
        prob.set_val("y", 5.0 * np.ones(size))

        prob.run_model()

        # Deriv should be 75. Analytic was wrong before the fix.
        data = prob.check_totals(method="cs", out_stream=None)

        assert_check_totals(data, atol=1e-6, rtol=1e-6)

    def test_multi_conn_inputs_manual_connect(self):

        prob = om.Problem()
        prob.model.add_subsystem('px1', om.IndepVarComp('x', 1.0))
        sub1 = prob.model.add_subsystem('sub1', om.Group())
        sub2 = prob.model.add_subsystem('sub2', om.Group())

        sub1.add_subsystem('src', om.ExecComp('y=x'))
        sub2.add_subsystem('comp1', om.ExecComp('z=x+y'))
        sub2.add_subsystem('comp2', om.ExecComp('z=x+y'))
        sub2.add_subsystem('comp3', om.ExecComp('z=x+y'))

        prob.model.connect('px1.x', 'sub1.src.x')
        prob.model.connect('sub1.src.y', 'sub2.comp1.y')
        prob.model.connect('sub1.src.y', 'sub2.comp2.y')
        prob.model.connect('sub1.src.y', 'sub2.comp3.y')

        sub2.approx_totals(method='cs')

        wrt = ['px1.x']
        of = ['sub2.comp1.z', 'sub2.comp2.z', 'sub2.comp3.z']

        prob.setup(mode='fwd')
        prob.run_model()

        assert_near_equal(prob['sub1.src.y'], 1.0, 1e-6)
        assert_near_equal(prob['sub2.comp1.z'], 2.0, 1e-6)
        assert_near_equal(prob['sub2.comp2.z'], 2.0, 1e-6)
        assert_near_equal(prob['sub2.comp3.z'], 2.0, 1e-6)

        data = prob.check_totals(of=of, wrt=wrt, method="fd", out_stream=None)
        assert_check_totals(data, atol=1e-6, rtol=1e-6)

        # Check the total derivatives in reverse mode
        prob.setup(mode='rev')
        prob.run_model()

        assert_near_equal(prob['sub1.src.y'], 1.0, 1e-6)
        assert_near_equal(prob['sub2.comp1.z'], 2.0, 1e-6)
        assert_near_equal(prob['sub2.comp2.z'], 2.0, 1e-6)
        assert_near_equal(prob['sub2.comp3.z'], 2.0, 1e-6)

        data = prob.check_totals(of=of, wrt=wrt, method="fd", out_stream=None)
        assert_check_totals(data, atol=1e-6, rtol=1e-6)

    def test_multi_conn_inputs_promoted(self):

        prob = om.Problem()
        prob.model.add_subsystem('px1', om.IndepVarComp('x', 1.0))
        sub1 = prob.model.add_subsystem('sub1', om.Group(), promotes=['y'])
        sub2 = prob.model.add_subsystem('sub2', om.Group(), promotes=['y'])

        sub1.add_subsystem('src', om.ExecComp('y=x'), promotes=['y'])
        sub2.add_subsystem('comp1', om.ExecComp('z=x+y'), promotes_inputs=['y'])
        sub2.add_subsystem('comp2', om.ExecComp('z=x+y'), promotes_inputs=['y'])
        sub2.add_subsystem('comp3', om.ExecComp('z=x+y'), promotes_inputs=['y'])

        prob.model.connect('px1.x', 'sub1.src.x')

        sub2.approx_totals(method='cs')

        wrt = ['px1.x']
        of = ['sub2.comp1.z', 'sub2.comp2.z', 'sub2.comp3.z']

        prob.setup(mode='fwd')
        prob.run_model()

        assert_near_equal(prob['y'], 1.0, 1e-6)
        assert_near_equal(prob['sub2.comp1.z'], 2.0, 1e-6)
        assert_near_equal(prob['sub2.comp2.z'], 2.0, 1e-6)
        assert_near_equal(prob['sub2.comp3.z'], 2.0, 1e-6)

        data = prob.check_totals(of=of, wrt=wrt, method="fd", out_stream=None)
        assert_check_totals(data, atol=1e-6, rtol=1e-6)

        # Check the total derivatives in reverse mode
        prob.setup(mode='rev')
        prob.run_model()

        assert_near_equal(prob['y'], 1.0, 1e-6)
        assert_near_equal(prob['sub2.comp1.z'], 2.0, 1e-6)
        assert_near_equal(prob['sub2.comp2.z'], 2.0, 1e-6)
        assert_near_equal(prob['sub2.comp3.z'], 2.0, 1e-6)

        data = prob.check_totals(of=of, wrt=wrt, method="fd", out_stream=None)
        assert_check_totals(data, atol=1e-6, rtol=1e-6)


class FakeGeomComp(om.ExplicitComponent):

    def initialize(self):
        self.options.declare("n", types=int)
        self.options.declare('declare_partials', types=(bool,), default=True)

    def setup(self):
        n = self.options["n"]
        self.add_input("x0", val=np.zeros(n), units="m")
        self.add_input("feather", val=0.0, units="deg")
        self.add_output("x", val=np.zeros(n), units="m")

        if self.options['declare_partials']:
            self.declare_partials("*", "*", method="fd")

        self._counter = 0

    def compute(self, inputs, outputs):
        self._counter += 1
        feather_rad = inputs["feather"][0]*np.pi/180
        x0 = inputs["x0"]

        outputs["x"][:] = 3*np.sin(feather_rad + 0.2)*x0 + 3*feather_rad**2


class FakeAeroComp(om.ExplicitComponent):

    def initialize(self):
        self.options.declare("n", types=int)
        self.options.declare('declare_partials', types=(bool,), default=True)

    def setup(self):
        n = self.options["n"]

        self.add_input("x", val=np.arange(n), units="m")
        self.add_input("omega", val=7000*2*np.pi/60, units="rad/s")

        self.add_output("CT", val=0.5)
        self.add_output("CP", val=0.5)

        if self.options['declare_partials']:
            self.declare_partials("*", "*", method="fd")

        self._counter = 0

    def compute(self, inputs, outputs):
        self._counter += 1
        omega = inputs["omega"][0]
        x = inputs["x"]

        outputs["CT"][0] = 0.8*omega**2 + np.sum(x)
        outputs["CP"][0] = 0.1*omega**3 + np.sum(x**2)


class GeometryAndAero2(om.Group):

    def initialize(self):
        self.options.declare("n", types=int)
        self.options.declare("rho", types=float)
        self.options.declare("vinf", types=float)
        self.options.declare('declare_partials', types=(bool,), default=True)

    def setup(self):
        n = self.options["n"]
        dp = self.options['declare_partials']

        comp = self.add_subsystem("init_geom", om.IndepVarComp(), promotes_outputs=["x0"])
        comp.add_output("x0", val=np.arange(n) + 1.0, units="m")

        self.add_subsystem("geom", FakeGeomComp(n=n, declare_partials=dp),
                           promotes_inputs=["x0", "feather"], promotes_outputs=["x"])
        self.add_subsystem("aero", FakeAeroComp(n=n, declare_partials=dp),
                           promotes_inputs=["x", "omega"], promotes_outputs=["CT", "CP"])

    def reset_count(self):
        self.geom._counter = 0
        self.aero._counter = 0


class TestSemiTotalsNumCalls(unittest.TestCase):

    def test_call_counts(self):
        size = 10
        rho = 1.17573
        minf = 0.111078231621482
        speedofsound = 344.5760217432
        vinf = minf*speedofsound

        prob = om.Problem()

        omega = 7199.759242*2*np.pi/60
        ivc = prob.model.add_subsystem("ivc", om.IndepVarComp(), promotes_outputs=["*"])
        ivc.add_output("feather", val=0.0, units="deg")
        ivc.add_output("omega", val=omega, units="rad/s")

        geom_and_aero = prob.model.add_subsystem('geom_and_aero',
                                                GeometryAndAero2(n=size, rho=rho, vinf=vinf),
                                                promotes_inputs=["feather", "omega"],
                                                promotes_outputs=["CT", "CP"])
        geom_and_aero.approx_totals(step=step, step_calc="abs", method="fd", form="forward")

        prob.model.add_design_var("feather", lower=-5.0, upper=25.0, units="deg", ref=1.0)
        prob.model.add_design_var("omega", lower=3000*2*np.pi/60, upper=7500*2*np.pi/60, units="rad/s", ref=1.0)
        prob.model.add_objective("CP", ref=1e0)

        prob.setup(force_alloc_complex=False)

        omega = (6245.096992023524*2*np.pi/60) + step
        feather = 0.6362159381168669
        prob.set_val("omega", omega, units="rad/s")
        prob.set_val("feather", feather, units="deg")
        prob.run_model()
        geom_and_aero.reset_count()
        prob.compute_totals(of=["CT"], wrt=["feather", "omega"])

        self.assertEqual(geom_and_aero.geom._counter, 2)
        self.assertEqual(geom_and_aero.aero._counter, 2)

        geom_and_aero.reset_count()
        data = prob.check_totals(method="fd", form="forward", step=step, step_calc="abs", out_stream=None)
        assert_check_totals(data, atol=1e-6, rtol=1e-6)

        self.assertEqual(geom_and_aero.geom._counter, 3)
        self.assertEqual(geom_and_aero.aero._counter, 4)

    def test_check_relevance_approx_totals(self):

            size = 10
            rho = 1.17573
            minf = 0.111078231621482
            speedofsound = 344.5760217432
            vinf = minf*speedofsound

            prob = om.Problem()
            prob.driver = om.ScipyOptimizeDriver()
            prob.driver.options["optimizer"] = "SLSQP"

            omega = 7199.759242*2*np.pi/60
            ivc = prob.model.add_subsystem("ivc", om.IndepVarComp(), promotes_outputs=["*"])
            ivc.add_output("feather", val=0.0, units="deg")
            ivc.add_output("omega", val=omega, units="rad/s")

            geom_and_aero = prob.model.add_subsystem('geom_and_aero',
                                                    GeometryAndAero2(n=size, rho=rho, vinf=vinf, declare_partials=False),
                                                    promotes_inputs=["feather", "omega"],
                                                    promotes_outputs=["CT", "CP"])
            geom_and_aero.approx_totals(step=step, step_calc="abs", method="fd", form="forward")

            prob.model.add_design_var("feather", lower=-5.0, upper=25.0, units="deg", ref=1.0)
            prob.model.add_design_var("omega", lower=3000*2*np.pi/60, upper=7500*2*np.pi/60, units="rad/s", ref=1.0)
            prob.model.add_objective("CP", ref=1e0)
            prob.model.add_constraint("CT", lower=0.0)

            prob.setup(force_alloc_complex=False)

            omega = (6245.096992023524*2*np.pi/60) + step
            feather = 0.6362159381168669
            prob.set_val("omega", omega, units="rad/s")
            prob.set_val("feather", feather, units="deg")
            fail = prob.run_driver()

            self.assertEqual(fail, False)


if __name__ == '__main__':
    unittest.main()



"""
A collection of utilities for file wrapping.

Note: This is a work in progress.
"""


import re

try:
    import pyparsing
    from pyparsing import CaselessLiteral, Combine, OneOrMore, Optional, \
        TokenConverter, Word, nums, oneOf, printables, ParserElement, alphanums
except ImportError:
    pyparsing = None
    TokenConverter = object

import numpy as np


def _getformat(val):
    """
    Get the output format for a floating point number.

    The general format is used with 16 places of accuracy, except for when
    the floating point value is an integer, in which case a decimal point
    followed by a single zero is used.

    Parameters
    ----------
    val : float or int
        the number which needs formatted.

    Returns
    -------
    string
        the format string.
    """
    if int(val) == val:
        return "%.1f"
    else:
        return "%.16g"


class _SubHelper(object):
    """
    Replaces file text at the correct word location in a line.

    This class contains the Helper Function that is passed to re.sub.

    Attributes
    ----------
    _newtext : str
        text to insert.
    _replace_location : int
        location in the file where replacement is to occur.
    _current_location : int
        current location in the file.
    _counter : int
        counter
    _start_location : int
        initial location where replacement is to occur.
    _end_location : int
        final location where replacement is to occur.
    """

    def __init__(self):
        """
        Initialize attributes.
        """
        self._newtext = ""
        self._replace_location = 0
        self._current_location = 0
        self._counter = 0
        self._start_location = 0
        self._end_location = 0

    def set(self, newtext, location):
        """
        Set a new word location and value for replacement.

        Parameters
        ----------
        newtext : str
            text to insert.
        location : int
            location in the file where replacement is to occur.
        """
        self._newtext = newtext
        self._replace_location = location
        self._current_location = 0

    def set_array(self, newtext, start_location, end_location):
        """
        Set a new starting location, ending location, and value for replacement.

        Parameters
        ----------
        newtext : str
            text to insert.
        start_location : int
            location
        end_location : int
            location
        """
        self._newtext = newtext
        self._start_location = start_location
        self._end_location = end_location
        self._current_location = 0

    def replace(self, text):
        """
        Replace text in file.

        This function should be passed to re.sub.

        Parameters
        ----------
        text : str
            text to insert.

        Returns
        -------
        string
            newtext if current location is replace location else the input text.
        """
        self._current_location += 1

        if self._current_location == self._replace_location:
            if isinstance(self._newtext, float):
                return _getformat(self._newtext) % self._newtext
            else:
                return str(self._newtext)
        else:
            return text.group()

    def replace_array(self, text):
        """
        Replace array of text values in file.

        This function should be passed to re.sub.

        Parameters
        ----------
        text : str
            text to insert.

        Returns
        -------
        string
            newtext if current location is replace location else the input text.
        """
        self._current_location += 1
        end = len(self._newtext)

        if self._current_location >= self._start_location and \
           self._current_location <= self._end_location and \
           self._counter < end:
            if isinstance(self._newtext[self._counter], float):
                val = self._newtext[self._counter]
                newval = _getformat(val) % val
            else:
                newval = str(self._newtext[self._counter])
            self._counter += 1
            return newval
        else:
            return text.group()


class _ToInteger(TokenConverter):
    """
    Converter for PyParsing that is used to turn a token into an int.
    """

    def postParse(self, instring, loc, tokenlist):
        """
        Convert token into an integer.

        Parameters
        ----------
        instring : str
            the input string
        loc : int
            the location of the matching string
        tokenlist : list
            list of matched tokens

        Returns
        -------
        int
            integer value for token.
        """
        return int(tokenlist[0])


class _ToFloat(TokenConverter):
    """
    Converter for PyParsing that is used to turn a token into a float.
    """

    def postParse(self, instring, loc, tokenlist):
        """
        Convert token into a float.

        Parameters
        ----------
        instring : str
            the input string
        loc : int
            the location of the matching string
        tokenlist : list
            list of matched tokens

        Returns
        -------
        float
            float value for token.
        """
        return float(tokenlist[0].replace('D', 'E'))


class _ToNan(TokenConverter):
    """
    Converter for PyParsing that is used to turn a token into Python nan.
    """

    def postParse(self, instring, loc, tokenlist):
        """
        Convert token into Python nan.

        Parameters
        ----------
        instring : str
            the input string
        loc : int
            the location of the matching string
        tokenlist : list
            list of matched tokens

        Returns
        -------
        float
            the float value for NaN.
        """
        return float('nan')


class _ToInf(TokenConverter):
    """
    Converter for PyParsing that is used to turn a token into Python inf.
    """

    def postParse(self, instring, loc, tokenlist):
        """
        Convert token into Python inf.

        Parameters
        ----------
        instring : str
            the input string
        loc : int
            the location of the matching string
        tokenlist : list
            list of matched tokens

        Returns
        -------
        float
            the float value for infinity.
        """
        return float('inf')


class InputFileGenerator(object):
    """
    Utility to generate an input file from a template.

    Substitution of values is supported. Data is located with a simple API.

    Attributes
    ----------
    _template_filename : str or None
        the name of the template file.
    _output_filename : str or None
        the name of the output file.
    _delimiter : int
        delimiter.
    _reg : int
        regular expression.
    _data : list of string
        the contents of the file, by line
    _current_row : int
        the current row of the file
    _anchored : bool
        indicator that position is relative to a landmark location.
    """

    def __init__(self):
        """
        Initialize attributes.
        """
        if pyparsing is None:
            raise RuntimeError("The 'pyparsing' package is required to use the file wrapping "
                               "utilities but it is not installed. Try 'pip install pyparsing'.")

        self._template_filename = None
        self._output_filename = None

        self._delimiter = " "
        self._reg = re.compile('[^ \n]+')

        self._data = []
        self._current_row = 0
        self._anchored = False

    def set_template_file(self, filename):
        """
        Set the name of the template file to be used.

        The template file is also read into memory when this method is called.

        Parameters
        ----------
        filename : str
            Name of the template file to be used.
        """
        self._template_filename = filename

        templatefile = open(filename, 'r')
        self._data = templatefile.readlines()
        templatefile.close()

    def set_generated_file(self, filename):
        """
        Set the name of the file that will be generated.

        Parameters
        ----------
        filename : str
            Name of the input file to be generated.
        """
        self._output_filename = filename

    def set_delimiters(self, delimiter):
        """
        Set the delimiters that are used to identify field boundaries.

        Parameters
        ----------
        delimiter : str
            A string containing characters to be used as delimiters.
        """
        self._delimiter = delimiter
        self._reg = re.compile('[^' + delimiter + '\n]+')

    def mark_anchor(self, anchor, occurrence=1):
        """
        Mark the location of a landmark.

        This lets you describe data by relative position. Note that a forward
        search begins at the old anchor location. If you want to restart the
        search for the anchor at the file beginning, then call ``reset_anchor()``
        before ``mark_anchor``.

        Parameters
        ----------
        anchor : str
            The text you want to search for.

        occurrence : int, optional
            Find nth instance of text; default is 1 (first). Use -1 to
            find last occurrence. Reverse searches always start at the end
            of the file no matter the state of any previous anchor.
        """
        if not isinstance(occurrence, int):
            raise ValueError("The value for occurrence must be an integer")

        instance = 0
        if occurrence > 0:
            count = 0
            max_lines = len(self._data)
            for index in range(self._current_row, max_lines):
                line = self._data[index]

                # If we are marking a new anchor from an existing anchor, and
                # the anchor is mid-line, then we still search the line, but
                # only after the anchor.
                if count == 0 and self._anchored:
                    line = line.split(anchor)[-1]

                if line.find(anchor) > -1:

                    instance += 1
                    if instance == occurrence:
                        self._current_row += count
                        self._anchored = True
                        return

                count += 1

        elif occurrence < 0:
            max_lines = len(self._data) - 1
            count = max_lines
            for index in range(max_lines, -1, -1):
                line = self._data[index]

                # If we are marking a new anchor from an existing anchor, and
                # the anchor is mid-line, then we still search the line, but
                # only before the anchor.
                if count == max_lines and self._anchored:
                    line = line.split(anchor)[0]

                if line.find(anchor) > -1:
                    instance += -1
                    if instance == occurrence:
                        self._current_row = count
                        self._anchored = True
                        return

                count -= 1
        else:
            raise ValueError("0 is not valid for an anchor occurrence.")

        raise RuntimeError("Could not find pattern %s in template file %s" %
                           (anchor, self._template_filename))

    def reset_anchor(self):
        """
        Reset anchor to the beginning of the file.
        """
        self._current_row = 0
        self._anchored = False

    def transfer_var(self, value, row, field):
        """
        Change a single variable in the template relative to the current anchor.

        Parameters
        ----------
        value : float, int, bool, str
            New value to set at the location.
        row : int
            Number of lines offset from anchor line (0 is anchor line).
            This can be negative.
        field : int
            Which word in line to replace, as denoted by delimiter(s).
        """
        j = self._current_row + row
        line = self._data[j]

        sub = _SubHelper()
        sub.set(value, field)
        newline = re.sub(self._reg, sub.replace, line)

        self._data[j] = newline

    def transfer_array(self, value, row_start, field_start, field_end,
                       row_end=None, sep=", "):
        """
        Change the values of an array in the template relative to the current anchor.

        This should generally be used for one-dimensional or free form arrays.

        Parameters
        ----------
        value : float, int, bool, str
            Array of values to insert.
        row_start : int
            Starting row for inserting the array. This is relative
            to the anchor, and can be negative.
        field_start : int
            Starting field in the given row_start as denoted by
            delimiter(s).
        field_end : int
            The final field the array uses in row_end.
            We need this to figure out if the template is too small or large.
        row_end : int, optional
            Use if the array wraps to cover additional lines.
        sep : int, optional
            Separator to use if we go beyond the template.
        """
        # Simplified input for single-line arrays
        if row_end is None:
            row_end = row_start

        sub = _SubHelper()

        for row in range(row_start, row_end + 1):
            j = self._current_row + row
            line = self._data[j]

            if row == row_end:
                f_end = field_end
            else:
                f_end = 99999

            sub.set_array(value, field_start, f_end)
            field_start = 0

            newline = re.sub(self._reg, sub.replace_array, line)
            self._data[j] = newline

        # Sometimes an array is too large for the example in the template
        # This is resolved by adding more fields at the end
        if sub._counter < len(value):
            for val in value[sub._counter:]:
                newline = newline.rstrip() + sep + str(val)
            self._data[j] = newline

        # Sometimes an array is too small for the template
        # This is resolved by removing fields
        elif sub._counter > len(value):
            # TODO - Figure out how to handle this.
            # Ideally, we'd remove the extra field placeholders
            raise ValueError("Array is too small for the template.")

    def transfer_2Darray(self, value, row_start, row_end, field_start, field_end):
        """
        Change the values of a 2D array in the template relative to the current anchor.

        This method is specialized for 2D arrays, where each row of the array is
        on its own line.

        Parameters
        ----------
        value : ndarray
            Array of values to insert.
        row_start : int
            Starting row for inserting the array. This is relative
            to the anchor, and can be negative.
        row_end : int
            Final row for the array, relative to the anchor.
        field_start : int
            Starting field in the given row_start as denoted by
            delimiter(s).
        field_end : int
            The final field the array uses in row_end.
            We need this to figure out if the template is too small or large.
        """
        sub = _SubHelper()

        i = 0

        for row in range(row_start, row_end + 1):
            j = self._current_row + row
            line = self._data[j]

            sub.set_array(value[i, :], field_start, field_end)

            newline = re.sub(self._reg, sub.replace_array, line)
            self._data[j] = newline

            sub._current_location = 0
            sub._counter = 0
            i += 1

        # TODO - Note, we currently can't handle going beyond the end of
        #        the template line

    def clearline(self, row):
        """
        Replace the contents of a row with the newline character.

        Parameters
        ----------
        row : int
            Row number to clear, relative to current anchor.
        """
        self._data[self._current_row + row] = "\n"

    def generate(self, return_data=False):
        """
        Use the template file to generate the input file.

        Parameters
        ----------
        return_data : bool
            If True, generated file data will be returned as a string.

        Returns
        -------
        string
            The generated file data if return_data is True or output filename
            has not been provided, else None.
        """
        if self._output_filename:
            with open(self._output_filename, 'w') as f:
                f.writelines(self._data)
        else:
            return_data = True

        if return_data:
            return '\n'.join(self._data)
        else:
            return None


class FileParser(object):
    """
    Utility to locate and read data from a file.

    Parameters
    ----------
    end_of_line_comment_char : str, optional
        End-of-line comment character to be ignored
        (e.g., Python supports in-line comments with "#").

    full_line_comment_char : str, optional
        Comment character that signifies a line should be skipped.

    Attributes
    ----------
    _filename : str
        the name of the file.
    _data : list of string
        the contents of the file, by line
    _delimiter : str
        the name of the file.
    _end_of_line_comment_char : str
        end-of-line comment character to be ignored.
    _full_line_comment_char : str
        comment character that signifies a line should be skipped.
    _current_row : int
        the current row of the file.
    _anchored : bool
        indicator that position is relative to a landmark location.
    """

    def __init__(self, end_of_line_comment_char=None, full_line_comment_char=None):
        """
        Initialize attributes.
        """
        if pyparsing is None:
            raise RuntimeError("The 'pyparsing' package is required to use the file wrapping "
                               "utilities but it is not installed. Try 'pip install pyparsing'.")

        self._filename = None
        self._data = []

        self._delimiter = " \t"
        self._end_of_line_comment_char = end_of_line_comment_char
        self._full_line_comment_char = full_line_comment_char

        self._current_row = 0
        self._anchored = False

        self.set_delimiters(self._delimiter)

    def set_file(self, filename):
        """
        Set the name of the file that will be generated.

        Parameters
        ----------
        filename : str
            Name of the input file to be generated.
        """
        self._filename = filename

        inputfile = open(filename, 'r')

        if not self._end_of_line_comment_char and not self._full_line_comment_char:
            self._data = inputfile.readlines()
        else:
            self._data = []
            for line in inputfile:
                if line[0] == self._full_line_comment_char:
                    continue
                self._data.append(line.split(self._end_of_line_comment_char)[0])

        inputfile.close()

    def set_delimiters(self, delimiter):
        r"""
        Set the delimiters that are used to identify field boundaries.

        Parameters
        ----------
        delimiter : str
            A string containing characters to be used as delimiters. The
            default value is ' \t', which means that spaces and tabs are not
            taken as data but instead mark the boundaries. Note that the
            parser is smart enough to recognize characters within quotes as
            non-delimiters.
        """
        self._delimiter = delimiter

        if delimiter != "columns":
            ParserElement.setDefaultWhitespaceChars(str(delimiter))

        self._reset_tokens()

    def mark_anchor(self, anchor, occurrence=1):
        """
        Mark the location of a landmark, which lets you describe data by relative position.

        Note that a forward search begins at the old anchor location. If you want to restart
        the search for the anchor at the file beginning, then call ``reset_anchor()`` before
        ``mark_anchor``.

        Parameters
        ----------
        anchor : str
            The text you want to search for.
        occurrence : int
            Find nth instance of text; default is 1 (first). Use -1 to
            find last occurrence. Reverse searches always start at the end
            of the file no matter the state of any previous anchor.
        """
        if not isinstance(occurrence, int):
            raise ValueError("The value for occurrence must be an integer")

        instance = 0

        if occurrence > 0:
            count = 0
            max_lines = len(self._data)
            for index in range(self._current_row, max_lines):
                line = self._data[index]

                # If we are marking a new anchor from an existing anchor, and
                # the anchor is mid-line, then we still search the line, but
                # only after the anchor.
                if count == 0 and self._anchored:
                    line = line.split(anchor)[-1]

                if anchor in line:

                    instance += 1
                    if instance == occurrence:
                        self._current_row += count
                        self._anchored = True
                        return

                count += 1

        elif occurrence < 0:
            max_lines = len(self._data) - 1
            count = max_lines
            for index in range(max_lines, -1, -1):
                line = self._data[index]

                # If we are marking a new anchor from an existing anchor, and
                # the anchor is mid-line, then we still search the line, but
                # only before the anchor.
                if count == max_lines and self._anchored:
                    line = line.split(anchor)[0]

                if anchor in line:
                    instance += -1
                    if instance == occurrence:
                        self._current_row = count
                        self._anchored = True
                        return

                count -= 1
        else:
            raise ValueError("0 is not valid for an anchor occurrence.")

        raise RuntimeError("Could not find pattern %s in output file %s" %
                           (anchor, self._filename))

    def reset_anchor(self):
        """
        Reset anchor to the beginning of the file.
        """
        self._current_row = 0
        self._anchored = False

    def transfer_line(self, row):
        """
        Return an entire line, relative to current anchor.

        Parameters
        ----------
        row : int
            Number of lines offset from anchor line (0 is anchor line).
            This can be negative.

        Returns
        -------
        string
            Line at the location requested.
        """
        return self._data[self._current_row + row].rstrip()

    def transfer_var(self, row, field, fieldend=None):
        """
        Get a single variable relative to the current anchor.

        Parameters
        ----------
        row : int
            Number of lines offset from anchor line (0 is anchor line).
            This can be negative.
        field : int
            If the delimiter is a set of chars: which word in line to retrieve.
            If the delimiter is 'columns': character position to start.
        fieldend : int (optional)
            If the delimiter is a set of chars: IGNORED.
            If the delimiter is 'columns': position of last character to return, or if
            omitted, the end of the line is used.

        Returns
        -------
        string
            Data from the requested location in the file.
        """
        j = self._current_row + row

        line = self._data[j]

        if self._delimiter == "columns":
            if not fieldend:
                line = line[(field - 1):]
            else:
                line = line[(field - 1):(fieldend)]

            # Let pyparsing figure out if this is a number, and return it
            # as a float or int as appropriate
            data = self._parse_line().parseString(line)

            # data might have been split if it contains whitespace. If so,
            # just return the whole string
            if len(data) > 1:
                return line
            else:
                return data[0]
        else:
            data = self._parse_line().parseString(line)
            return data[field - 1]

    def transfer_keyvar(self, key, field, occurrence=1, rowoffset=0):
        """
        Search for a key relative to the current anchor and get a field from that line.

        You can do the same thing with a call to ``mark_anchor`` and ``transfer_var``.
        This function just combines them for convenience.

        Parameters
        ----------
        key : str
            The key to search for.
        field : int
            Which field to transfer. Field 0 is the key.
        occurrence : int
            Find nth instance of text; default is 1 (first value
            field). Use -1 to find last occurance. Position 0 is the key
            field, so it should not be used as a value for occurrence.
        rowoffset : int (optional)
            Optional row offset from the occurrence of key. This can
            also be negative.

        Returns
        -------
        string
            Data from the requested location in the file.
        """
        if not isinstance(occurrence, int) or occurrence == 0:
            msg = "The value for occurrence must be a nonzero integer"
            raise ValueError(msg)

        instance = 0
        if occurrence > 0:
            row = 0
            for line in self._data[self._current_row:]:
                if line.find(key) > -1:
                    instance += 1
                    if instance == occurrence:
                        break
                row += 1

        elif occurrence < 0:
            row = -1
            for line in reversed(self._data[self._current_row:]):
                if line.find(key) > -1:
                    instance += -1
                    if instance == occurrence:
                        break
                row -= 1

        j = self._current_row + row + rowoffset
        line = self._data[j]

        fields = self._parse_line().parseString(line.replace(key, "KeyField"))

        return fields[field]

    def transfer_array(self, rowstart, fieldstart, rowend=None, fieldend=None):
        """
        Get an array of variables relative to the current anchor.

        Setting the delimiter to 'columns' elicits some special behavior
        from this method. Normally, the extraction process wraps around
        at the end of a line and continues grabbing each field at the start of
        a newline. When the delimiter is set to columns, the parameters
        (rowstart, fieldstart, rowend, fieldend) demark a box, and all
        values in that box are retrieved. Note that standard whitespace
        is the secondary delimiter in this case.

        Parameters
        ----------
        rowstart : int
            Row number to start, relative to the current anchor.
        fieldstart : int
            Field number to start.
        rowend : int, optional
            Row number to end. If not set, then only one row is grabbed.
        fieldend : int
            Field number to end.

        Returns
        -------
        string
            Data from the requested location in the file.
        """
        j1 = self._current_row + rowstart

        if rowend is None:
            j2 = j1 + 1
        else:
            j2 = self._current_row + rowend + 1

        if not fieldend:
            raise ValueError("fieldend is missing, currently required")

        lines = self._data[j1:j2]

        data = np.zeros(shape=(0, 0))

        for i, line in enumerate(lines):
            if self._delimiter == "columns":
                line = line[(fieldstart - 1):fieldend]

                # Stripping whitespace may be controversial.
                line = line.strip()

                # Let pyparsing figure out if this is a number, and return it
                # as a float or int as appropriate
                parsed = self._parse_line().parseString(line)

                newdata = np.array(parsed[:])
                # data might have been split if it contains whitespace. If the
                # data is string, we probably didn't want this.
                if newdata.dtype.type is np.str_:
                    newdata = np.array(line)

                data = np.append(data, newdata)
            else:
                parsed = self._parse_line().parseString(line)

                if i == j2 - j1 - 1:
                    data = np.append(data, np.array(parsed[(fieldstart - 1):fieldend]))
                else:
                    data = np.append(data, np.array(parsed[(fieldstart - 1):]))

                fieldstart = 1

        return data

    def transfer_2Darray(self, rowstart, fieldstart, rowend, fieldend=None):
        """
        Get a 2D array of variables relative to the current anchor.

        Each line of data is placed in a separate row.

        If the delimiter is set to 'columns', then the values contained in
        fieldstart and fieldend should be the column number instead of the
        field number.

        Parameters
        ----------
        rowstart : int
            Row number to start, relative to the current anchor.
        fieldstart : int
            Field number to start.
        rowend : int
            Row number to end relative to current anchor.
        fieldend : int (optional)
            Field number to end. If not specified, grabs all fields up to the
            end of the line.

        Returns
        -------
        string
            Data from the requested location in the file.
        """
        if fieldend and (fieldstart > fieldend):
            msg = "fieldend must be greater than fieldstart"
            raise ValueError(msg)

        if rowstart > rowend:
            msg = "rowend must be greater than rowstart"
            raise ValueError(msg)

        j1 = self._current_row + rowstart
        j2 = self._current_row + rowend + 1
        lines = list(self._data[j1:j2])

        if self._delimiter == "columns":
            if fieldend:
                line = lines[0][(fieldstart - 1):fieldend]
            else:
                line = lines[0][(fieldstart - 1):]

            parsed = self._parse_line().parseString(line)
            row = np.array(parsed[:])
            data = np.zeros(shape=(abs(j2 - j1), len(row)))
            data[0, :] = row

            for i, line in enumerate(list(lines[1:])):
                if fieldend:
                    line = line[(fieldstart - 1):fieldend]
                else:
                    line = line[(fieldstart - 1):]

                parsed = self._parse_line().parseString(line)
                data[i + 1, :] = np.array(parsed[:])
        else:
            parsed = self._parse_line().parseString(lines[0])
            if fieldend:
                row = np.array(parsed[(fieldstart - 1):fieldend])
            else:
                row = np.array(parsed[(fieldstart - 1):])

            data = np.zeros(shape=(abs(j2 - j1), len(row)))
            data[0, :] = row

            for i, line in enumerate(list(lines[1:])):
                parsed = self._parse_line().parseString(line)

                if fieldend:
                    try:
                        data[i + 1, :] = np.array(parsed[(fieldstart - 1):fieldend])
                    except Exception:
                        print(data)
                else:
                    data[i + 1, :] = np.array(parsed[(fieldstart - 1):])

        return data

    def _parse_line(self):
        """
        Parse a single data line that may contain string or numerical data.

        Float and Int 'words' are converted to their appropriate type.
        Exponentiation is supported, as are NaN and Inf.

        Returns
        -------
        <ParserElement>
            the parsed line.
        """
        return self.line_parse_token

    def _reset_tokens(self):
        """
        Set up the tokens for pyparsing.
        """
        # Somewhat of a hack, but we can only use printables if the delimiter is
        # just whitespace. Otherwise, some seprators (like ',' or '=') potentially
        # get parsed into the general string text. So, if we have non whitespace
        # delimiters, we need to fall back to just alphanums, and then add in any
        # missing but important symbols to parse.
        if self._delimiter.isspace():
            textchars = printables
        else:
            textchars = alphanums

            symbols = ['.', '/', '+', '*', '^', '(', ')', '[', ']', '=',
                       ':', ';', '?', '%', '&', '!', '#', '|', '<', '>',
                       '{', '}', '-', '_', '@', '$', '~']

            for symbol in symbols:
                if symbol not in self._delimiter:
                    textchars = textchars + symbol

        digits = Word(nums)
        dot = "."
        sign = oneOf("+ -")
        ee = CaselessLiteral('E') | CaselessLiteral('D')

        num_int = _ToInteger(Combine(Optional(sign) + digits))

        num_float = _ToFloat(Combine(
            Optional(sign) +
            ((digits + dot + Optional(digits)) | (dot + digits)) +
            Optional(ee + Optional(sign) + digits)
        ))

        # special case for a float written like "3e5"
        mixed_exp = _ToFloat(Combine(digits + ee + Optional(sign) + digits))

        nan = (_ToInf(oneOf("Inf -Inf")) |
               _ToNan(oneOf("NaN nan NaN%  NaNQ NaNS qNaN sNaN 1.#SNAN 1.#QNAN -1.#IND")))

        string_text = Word(textchars)

        self.line_parse_token = (OneOrMore((nan | num_float | mixed_exp | num_int | string_text)))



"""Define the NNBase class."""

import numpy as np

from math import ceil
from scipy.spatial import KDTree


class NNBase(object):
    """
    Base class for common functionality between nearest neighbor interpolants.

    Parameters
    ----------
    training_points : ndarray
        Ndarray of shape (num_points x independent dims) containing training input locations.
    training_values : ndarray
        Ndarray of shape (num_points x dependent dims) containing training output values.
    num_leaves : int
        How many leaves the tree should have.

    Attributes
    ----------
    _tpm : ndarray
        Ndarray of shape (1 x independent dims) containing the minimum in each dimension of
         the training input locations.
    _tpr : ndarray
        Ndarray of shape (1x independent dims) containing the range of each dimension of
        the training input locations.
    _tvm : ndarray
        Ndarray of shape (1 x independent dims) containing the minimum in each dimension of
        the training output values.
    _tvr : ndarray
        Ndarray of shape (1x independent dims) containing the range of each dimension of
        the training output values.
    _tp : ndarray
        Ndarray of shape (num_points x independent dims) containing normalized training
        input locations.
    _tv : ndarray
        Ndarray of shape (num_points x independent dims) containing normalized training
        output values.
    _indep_dims : int
        Number of independent dims
    _dep_dims : int
        Number of dependent dims
    _ntpts : int
        Number of training points
    _KData : scipy.spatial.cKDTree
        KDTree used for finding the nearest neighbors.
    _pt_cache : tuple(ndarray, ndarray, ndarray)
        Internal cache of the last found neighbors.
    """

    def __init__(self, training_points, training_values, num_leaves=2):
        """
        Initialize nearest neighbor interpolant by scaling input to the unit hypercube.
        """
        # training_points and training_values are the known points and their
        # respective values which will be interpolated against.
        # Grab the mins and ranges of each dimension
        self._tpm = np.amin(training_points, axis=0)
        self._tpr = (np.amax(training_points, axis=0) - self._tpm)
        self._tvm = np.amin(training_values, axis=0)
        self._tvr = (np.amax(training_values, axis=0) - self._tvm)

        # This prevents against collinear data (range = 0)
        self._tpr[self._tpr == 0] = 1
        self._tvr[self._tvr == 0] = 1

        # Normalize all points
        self._tp = (training_points - self._tpm) / self._tpr
        self._tv = (training_values - self._tvm) / self._tvr

        # Record number of dimensions and points
        self._indep_dims = training_points.shape[1]
        self._dep_dims = training_values.shape[1]
        self._ntpts = training_points.shape[0]

        # Make training data into a Tree
        leavesz = ceil(self._ntpts / float(num_leaves))
        self._KData = KDTree(self._tp, leafsize=leavesz)

        # Cache for gradients
        self._pt_cache = None



"""Base class used to define the interface for derivative approximation schemes."""
import time
from collections import defaultdict
from itertools import repeat
import numpy as np

from openmdao.core.constants import INT_DTYPE
from openmdao.vectors.vector import _full_slice
from openmdao.utils.array_utils import get_input_idx_split, ValueRepeater
import openmdao.utils.coloring as coloring_mod
from openmdao.utils.general_utils import _convert_auto_ivc_to_conn_name, LocalRangeIterable
from openmdao.utils.mpi import check_mpi_env
from openmdao.utils.rangemapper import RangeMapper


use_mpi = check_mpi_env()
if use_mpi is False:
    PETSc = None
else:
    try:
        from petsc4py import PETSc
    except ImportError:
        if use_mpi:
            raise ImportError("Importing petsc4py failed and OPENMDAO_USE_MPI is true.")
        PETSc = None


class ApproximationScheme(object):
    """
    Base class used to define the interface for derivative approximation schemes.

    Attributes
    ----------
    _approx_groups : list
        A list of approximation tuples ordered into groups of 'of's matching the same 'wrt'.
    _colored_approx_groups: list
        A list containing info for all colored approximation groups.
    _approx_groups_cached_under_cs : bool
        Flag indicates whether approx_groups was generated under complex step from higher in the
        model hieararchy.
    _wrt_meta : dict
        A dict that maps wrt name to its fd/cs metadata.
    _progress_out : None or file-like object
        Attribute to output the progress of check_totals
    _jac_scatter : tuple
        Data needed to scatter values from results array to a total jacobian column.
    _totals_directions : dict
        If directional total derivatives are being computed, this will contain the direction keyed
        by mode ('fwd' or 'rev').
    _totals_directional_mode : str or None
        If directional total derivatives are being computed, this will contain the top level
        mode ('fwd' or 'rev'), else None.
    """

    def __init__(self):
        """
        Initialize the ApproximationScheme.
        """
        self._approx_groups = None
        self._colored_approx_groups = None
        self._approx_groups_cached_under_cs = False
        self._wrt_meta = {}
        self._progress_out = None
        self._jac_scatter = None
        self._totals_directions = {}
        self._totals_directional_mode = None

    def __repr__(self):
        """
        Return a simple string representation.

        Returns
        -------
        str
            String containing class name and added approximation keys.
        """
        return f"at {id(self)}, {self.__class__.__name__}: {list(self._wrt_meta.keys())}"

    def _reset(self):
        """
        Get rid of any existing approx groups.
        """
        self._colored_approx_groups = None
        self._approx_groups = None

    def _get_approx_groups(self, system, under_cs=False):
        """
        Retrieve data structure that contains all the approximations.

        This data structure is regenerated if we transition to or from being under a complex step
        from higher in the model hierarchy.

        Parameters
        ----------
        system : <System>
            Group or component instance.
        under_cs : bool
            Flag that indicates if we are under complex step.

        Returns
        -------
        Tuple  (approx_groups, colored_approx_groups)
            Each approx_groups entry contains specific data for a wrt var.
            Each colored_approx_groups entry contains data for a group of columns.
        """
        if under_cs != self._approx_groups_cached_under_cs:
            if coloring_mod._use_partial_sparsity:
                self._init_colored_approximations(system)
            self._init_approximations(system)
        else:
            if self._colored_approx_groups is None and coloring_mod._use_partial_sparsity:
                self._init_colored_approximations(system)
            if self._approx_groups is None:
                self._init_approximations(system)

        self._approx_groups_cached_under_cs = under_cs

        return self._approx_groups, self._colored_approx_groups

    def add_approximation(self, abs_key, system, kwargs):
        """
        Use this approximation scheme to approximate the derivative d(of)/d(wrt).

        Parameters
        ----------
        abs_key : tuple(str,str)
            Absolute name pairing of (of, wrt) for the derivative.
        system : System
            Containing System.
        kwargs : dict
            Additional keyword arguments, to be interpreted by sub-classes.
        """
        raise NotImplementedError("add_approximation has not been implemented")

    def _init_colored_approximations(self, system):
        is_total = system.pathname == ''
        is_semi = _is_group(system) and not is_total
        self._colored_approx_groups = []
        wrt_ranges = []

        # don't do anything if the coloring doesn't exist yet, or if there is no
        # forward coloring
        coloring = system._coloring_info.coloring
        if not isinstance(coloring, coloring_mod.Coloring) or coloring._fwd is None:
            return

        wrt_matches = system._coloring_info._update_wrt_matches(system)
        out_slices = system._outputs.get_slice_dict()

        # this maps column indices into colored jac into indices into full jac
        if wrt_matches is not None:
            ccol2jcol = np.empty(coloring._shape[1], dtype=INT_DTYPE)

        # colored col to out vec idx
        if is_total:
            ccol2outvec = np.empty(coloring._shape[1], dtype=INT_DTYPE)

        if is_total or wrt_matches is not None:
            colored_start = colored_end = 0
            for abs_wrt, cstart, cend, _, cinds, _ in system._jac_wrt_iter():
                if wrt_matches is None or abs_wrt in wrt_matches:
                    colored_end += cend - cstart
                    if wrt_matches is not None:
                        ccol2jcol[colored_start:colored_end] = range(cstart, cend)
                    if is_total and abs_wrt in out_slices:
                        slc = out_slices[abs_wrt]
                        if cinds is not None:
                            rng = np.arange(slc.start, slc.stop)[cinds]
                        else:
                            rng = range(slc.start, slc.stop)
                        wrt_ranges.append((abs_wrt, slc.stop - slc.start))
                        ccol2outvec[colored_start:colored_end] = rng
                    colored_start = colored_end

        row_var_sizes = {v: sz for v, sz in zip(coloring._row_vars, coloring._row_var_sizes)}
        row_map = np.empty(coloring._shape[0], dtype=INT_DTYPE)
        abs2prom = system._var_allprocs_abs2prom['output']

        if is_total:
            it = ((of, end - start) for of, start, end, _, _ in system._jac_of_iter())
            rangemapper = RangeMapper.create(wrt_ranges)
        else:
            it = ((n, arr.size) for n, arr in system._outputs._abs_item_iter())

        start = end = colorstart = colorend = 0
        for name, sz in it:
            end += sz
            prom = name if is_total else abs2prom[name]
            if prom in row_var_sizes:
                colorend += row_var_sizes[prom]
                row_map[colorstart:colorend] = range(start, end)
                colorstart = colorend
            start = end

        for wrt, meta in self._wrt_meta.items():
            if wrt_matches is None or wrt in wrt_matches:
                # data is the same for all colored approxs so we only need the first
                data = self._get_approx_data(system, wrt, meta)
                break

        outputs = system._outputs
        inputs = system._inputs

        from openmdao.core.implicitcomponent import ImplicitComponent
        use_full_cols = is_semi or isinstance(system, ImplicitComponent)

        for cols, nzrows in coloring.color_nonzero_iter('fwd'):
            nzrows = [row_map[r] for r in nzrows]
            jaccols = cols if wrt_matches is None else ccol2jcol[cols]
            if is_total:
                vcols = ccol2outvec[cols]
                seed_vars = rangemapper.inds2keys(cols)
            else:
                vcols = jaccols
                seed_vars = None
            vec_ind_list = get_input_idx_split(vcols, inputs, outputs, use_full_cols, is_total)
            self._colored_approx_groups.append((data, jaccols, vec_ind_list, nzrows, seed_vars))

    def _init_approximations(self, system):
        """
        Prepare for later approximations.

        Parameters
        ----------
        system : System
            The system having its derivs approximated.
        """
        total = system.pathname == ''

        in_slices = system._inputs.get_slice_dict()
        out_slices = system._outputs.get_slice_dict()

        coloring = system._get_static_coloring()

        self._approx_groups = []
        self._nruns_uncolored = 0

        if system._during_sparsity:
            wrt_matches = system._coloring_info.wrt_matches
        else:
            wrt_matches = None

        if self._totals_directional_mode == 'rev':
            wrts_directional = []
            in_inds_directional = []
            vec_inds_directional = defaultdict(list)

        # wrt here is an absolute name (source if total)
        for wrt, start, end, vec, sinds, _ in system._jac_wrt_iter(wrt_matches):
            if wrt in self._wrt_meta:
                meta = self._wrt_meta[wrt]
                if coloring is not None and 'coloring' in meta:
                    continue
                if vec is system._inputs:
                    slices = in_slices
                else:
                    slices = out_slices

                data = self._get_approx_data(system, wrt, meta)
                directional = meta['directional'] or self._totals_directions

                in_idx = range(start, end)

                if total and sinds is not _full_slice:
                    if vec is None:
                        vec_idx = ValueRepeater(None, sinds.size)
                    else:
                        # local index into var
                        vec_idx = sinds.copy()
                        # convert into index into input or output vector
                        vec_idx += slices[wrt].start
                        # Directional derivatives for quick deriv checking.
                        # Place the indices in a list so that they are all stepped at the same time.
                        if directional:
                            in_idx = [list(in_idx)]
                            vec_idx = [vec_idx]
                else:
                    vec_idx = LocalRangeIterable(system, wrt)
                    if directional and vec is not None:
                        vec_idx = [v for v in vec_idx if v is not None]

                    # Directional derivatives for quick deriv checking.
                    # Place the indices in a list so that they are all stepped at the same time.
                    if directional:
                        in_idx = [list(in_idx)]
                        vec_idx = [list(vec_idx)]

                if directional:
                    self._nruns_uncolored += 1
                else:
                    self._nruns_uncolored += end - start

                if self._totals_directional_mode == 'rev':
                    wrts_directional.append(wrt)
                    data_directional = data
                    in_inds_directional.extend(in_idx[0])
                    vec_inds_directional[vec].extend(vec_idx[0])
                else:
                    if self._totals_directions:
                        direction = self._totals_directions['fwd'][start:end]
                    else:
                        direction = meta['vector']
                    self._approx_groups.append(((wrt,) if directional else wrt, data, in_idx,
                                                [(vec, vec_idx)], directional, direction))

        if total:
            if self._totals_directional_mode == 'rev':
                self._nruns_uncolored = 1
                vector = self._totals_directions['fwd']
                self._approx_groups = [(tuple(wrts_directional), data_directional,
                                        [in_inds_directional], list(vec_inds_directional.items()),
                                        True, vector)]

            # compute scatter from the results vector into a column of the total jacobian
            sinds, tinds, colsize, has_dist_data = system._get_jac_col_scatter()
            if has_dist_data:
                src_vec = PETSc.Vec().createWithArray(np.zeros(len(system._outputs), dtype=float),
                                                      comm=system.comm)
                tgt_vec = PETSc.Vec().createWithArray(np.zeros(colsize, dtype=float),
                                                      comm=system.comm)
                src_inds = PETSc.IS().createGeneral(sinds, comm=system.comm)
                tgt_inds = PETSc.IS().createGeneral(tinds, comm=system.comm)
                self._jac_scatter = (has_dist_data,
                                     PETSc.Scatter().create(src_vec, src_inds, tgt_vec, tgt_inds),
                                     src_vec, tgt_vec)
            else:
                self._jac_scatter = (has_dist_data, sinds, tinds)
        else:
            self._jac_scatter = None

    def _colored_column_iter(self, system, colored_approx_groups):
        """
        Perform colored approximations and yields (column_index, column) for each jac column.

        Parameters
        ----------
        system : System
            System where this approximation is occurring.
        colored_approx_groups : list of tuples of the form (data, jaccols, vec_ind_list, nzrows)
            data -> metadata needed to perform cs or fd
            jaccols -> jacobian columns corresponding to a colored solve
            vec_ind_list -> list of tuples of the form (Vector, ndarray of int)
                Tuple of wrt indices and corresponding data vector to perturb.
            nzrows -> rows containing nonzero values for each column in jaccols

        Yields
        ------
        int
            column index
        ndarray
            solution array corresponding to the jacobian column at the given column index
        """
        total = system.pathname == ''
        total_or_semi = total or _is_group(system)

        if total:
            tot_result = np.zeros(sum([end - start for _, start, end, _, _
                                       in system._jac_of_iter()]))
            scratch = tot_result.copy()
        else:
            scratch = np.empty(len(system._outputs))

        # Clean vector for results (copy of the outputs or resids)
        vec = system._outputs if total_or_semi else system._residuals
        results_array = vec.asarray(copy=True)

        use_parallel_fd = system._num_par_fd > 1 and (system._full_comm is not None and
                                                      system._full_comm.size > 1)
        num_par_fd = system._num_par_fd if use_parallel_fd else 1
        par_fd_w_serial_model = use_parallel_fd and system._num_par_fd == system._full_comm.size
        fd_count = 0
        mycomm = system._full_comm if use_parallel_fd else system.comm

        nruns = len(colored_approx_groups)
        tosend = None

        for data, jcols, vec_ind_list, nzrows, seed_vars, in colored_approx_groups:
            mult = self._get_multiplier(data)

            if fd_count % num_par_fd == system._par_fd_id:
                # run the finite difference
                with system._relevance.seeds_active(fwd_seeds=seed_vars):
                    result = self._run_point(system, vec_ind_list, data, results_array,
                                             total_or_semi)

                if par_fd_w_serial_model or not use_parallel_fd:
                    result = self._transform_result(result)

                    if mult != 1.0:
                        result *= mult

                    if total:
                        result = self._get_total_result(result, tot_result)

                    tosend = (fd_count, result)

                else:  # parallel model (some vars are remote)
                    raise NotImplementedError("simul approx coloring with parallel FD/CS is "
                                              "only supported currently when using "
                                              "a serial model, i.e., when "
                                              "num_par_fd == number of MPI procs.")

            fd_count += 1

            # check if it's time to collect parallel FD columns
            if use_parallel_fd and (nruns < num_par_fd or fd_count % num_par_fd == 0 or
                                    fd_count == nruns):
                allres = mycomm.allgather(tosend)
                tosend = None
            else:
                allres = [tosend]

            for tup in allres:
                if tup is None:
                    continue

                i, res = tup

                _, jcols, _, nzrows, _ = colored_approx_groups[i]

                for i, col in enumerate(jcols):
                    scratch[:] = 0.0
                    scratch[nzrows[i]] = res[nzrows[i]]
                    yield col, scratch

    def _vec_ind_iter(self, vec_ind_list):
        """
        Yield the vector index list as one chunk if doing directional totals.

        If not, yield each index individually, along with the vector to index into.

        Parameters
        ----------
        vec_ind_list : list
            List of (Vector, indices) tuples.

        Yields
        ------
        list
            [[Vector, inds]]
        """
        if self._totals_directions:
            yield vec_ind_list, vec_ind_list[0][1]
        else:
            entry = [[None, None]]
            ent0 = entry[0]
            for vec, vec_idxs in vec_ind_list:
                if vec_idxs is None:
                    continue
                for vinds in vec_idxs:
                    ent0[0] = vec
                    ent0[1] = vinds
                    yield entry, vinds

    def _uncolored_column_iter(self, system, approx_groups):
        """
        Perform approximations and yield (column_index, column) for each jac column.

        Parameters
        ----------
        system : System
            System where this approximation is occurring.
        approx_groups : list of tuples of the form (wrt, data, jaccols, vec, vec_idx, directional,
                                                    dir_vector)
            wrt -> name of the 'with respect to' variable
            data -> metadata needed to perform cs or fd
            jaccols -> jacobian columns corresponding to all solves for the 'wrt' variable
            vec -> Vector being perturbed
            vec_idx -> indices where the vector will be perturbed (one per approximation)
            directional -> if True we're computing a directional derivative (one approx for the
                           whole wrt variable instead of 1 per entry in the variable)
            dir_vector -> if directional is True, this may contain the direction vector

        Yields
        ------
        int
            column index
        ndarray
            solution array corresponding to the jacobian column at the given column index
        """
        total = system.pathname == ''
        if total:
            for _, _, end, _, _ in system._jac_of_iter():
                pass
            tot_result = np.zeros(end)

        total_or_semi = total or _is_group(system)

        # Clean vector for results (copy of the outputs or resids)
        results_array = system._outputs.asarray(copy=True) if total_or_semi \
            else system._residuals.asarray(copy=True)
        use_parallel_fd = system._num_par_fd > 1 and (system._full_comm is not None and
                                                      system._full_comm.size > 1)
        num_par_fd = system._num_par_fd if use_parallel_fd else 1

        nruns = self._nruns_uncolored
        tosend = None
        fd_count = 0
        mycomm = system._full_comm if use_parallel_fd else system.comm

        # now do uncolored solves
        for group_i, tup in enumerate(approx_groups):
            wrt, data, jcol_idxs, vec_ind_list, directional, direction = tup
            if self._progress_out:
                start_time = time.perf_counter()

            if direction is not None:
                app_data = self.apply_directional(data, direction)
            else:
                app_data = data

            mult = self._get_multiplier(data)

            jidx_iter = iter(range(len(jcol_idxs)))
            for vec_ind_info, vecidxs in self._vec_ind_iter(vec_ind_list):

                if fd_count % num_par_fd == system._par_fd_id:
                    # run the finite difference
                    if total:
                        seeds = wrt if directional else (wrt,)
                        with system._relevance.seeds_active(fwd_seeds=seeds):
                            result = self._run_point(system, vec_ind_info,
                                                     app_data, results_array, total_or_semi,
                                                     jcol_idxs)
                    else:
                        result = self._run_point(system, vec_ind_info,
                                                 app_data, results_array, total_or_semi,
                                                 jcol_idxs)

                    result = self._transform_result(result)

                    if direction is not None or mult != 1.0:
                        result *= mult

                    if total:
                        result = self._get_total_result(result, tot_result)

                    if vecidxs is None and not total_or_semi:
                        tosend = (group_i, None, None)
                    else:
                        tosend = (group_i, next(jidx_iter), result)  # use local jac row var index

                    if self._progress_out:
                        end_time = time.perf_counter()
                        prom_name = _convert_auto_ivc_to_conn_name(
                            system._conn_global_abs_in2out, wrt)
                        self._progress_out.write(f"{fd_count + 1}/{len(result)}: Checking "
                                                 f"derivatives with respect to: "
                                                 f"'{prom_name} [{vecidxs}]' ... "
                                                 f"{round(end_time - start_time, 4)} seconds\n")
                elif use_parallel_fd:
                    next(jidx_iter)  # skip this column index

                fd_count += 1

                # check if it's time to collect parallel FD columns
                if use_parallel_fd:
                    if fd_count == nruns or fd_count % num_par_fd == 0:
                        allres = mycomm.allgather(tosend)
                        tosend = None
                    else:
                        continue
                else:
                    allres = [tosend]

                for tup in allres:
                    if tup is None:
                        continue
                    gi, icount, res = tup
                    if icount is None:
                        continue  # skip yield for non-local partial jac column
                    # approx_groups[gi] -> (wrt, data, jcol_idxs, vec_ind_list, direction)
                    # [2] -> jcol_idxs, and [icount] -> actual indices used for the fd run.
                    jinds = approx_groups[gi][2][icount]
                    if directional:
                        yield jinds[0], res
                    else:
                        yield jinds, res

    def compute_approximations(self, system, jac=None):
        """
        Execute the system to compute the approximate sub-Jacobians.

        Parameters
        ----------
        system : System
            System on which the execution is run.
        jac : None or dict-like
            If None, update system with the approximated sub-Jacobians. Otherwise, store the
            approximations in the given dict-like object.
        """
        if not self._wrt_meta:
            return

        if system._tot_jac is not None:
            jac = system._tot_jac
        elif jac is None:
            jac = system._jacobian

        for ic, col in self.compute_approx_col_iter(system,
                                                    under_cs=system._outputs._under_complex_step):
            jac.set_col(system, ic, col)

    def _compute_approx_col_iter(self, system, under_cs):
        # This will either generate new approx groups or use cached ones
        approx_groups, colored_approx_groups = self._get_approx_groups(system, under_cs)

        if colored_approx_groups:
            yield from self._colored_column_iter(system, colored_approx_groups)

        yield from self._uncolored_column_iter(system, approx_groups)

    def _get_total_result(self, outarr, totarr):
        """
        Convert output array into a column array that matches the size of the total jacobian.

        Also gather any remote vars, if necessary, into the column array.

        Parameters
        ----------
        outarr : ndarray
            Array containing local results from the outputs vector.
        totarr : ndarray
            Array sized to fit a total jac column.

        Returns
        -------
        ndarray
            totarr, now filled with current values, potentially from other mpi procs.
        """
        if self._jac_scatter[0]:  # dist data passing
            _, scatter, svec, tvec = self._jac_scatter
            svec.array = outarr
            tvec.array[:] = totarr
            scatter.scatter(svec, tvec, addv=False, mode=False)
            totarr[:] = tvec.array
        else:
            _, sinds, tinds = self._jac_scatter
            totarr[tinds] = outarr[sinds]

        return totarr


def _is_group(obj):
    """
    Return True if the given object is Group.

    Parameters
    ----------
    obj : object
        Object to be checked.
    """
    from openmdao.core.group import Group
    return isinstance(obj, Group)



import unittest

from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.rangemapper import RangeMapper, RangeTree, FlatRangeMapper, MAX_FLAT_RANGE_SIZE


_data = {
    'a': 1,  # 0:1
    'b': 8,  # 1:9
    'x': 6,  # 9:15
    'y': 21, # 15:36
    'z': 6,  # 36:42
}


class TestRangeMapper(unittest.TestCase):
    def test_create(self):
        mapper = RangeMapper.create(_data.items())
        self.assertEqual(type(mapper), FlatRangeMapper)
        mapper = RangeMapper.create(_data.items(), max_flat_range_size=40)
        self.assertEqual(type(mapper), RangeTree)

    def test_get_item(self):
        for mclass in (RangeTree, FlatRangeMapper):
            with self.subTest(msg=f'{mclass.__name__} test'):
                mapper = mclass(_data.items())
                inds = [0, 1, 7, 9, 14, 15, 22, 41, 42, 43]
                expected = ['a', 'b', 'b', 'x', 'x', 'y', 'y', 'z', None, None]
                for i, ex_i in zip(inds, expected):
                    got = mapper[i]
                    self.assertEqual(got, ex_i)

    def test_key2range(self):
        for mclass in (RangeTree, FlatRangeMapper):
            with self.subTest(msg=f'{mclass.__name__} test'):
                mapper = mclass(_data.items())
                keys = ['a', 'b', 'x', 'y', 'z']
                expected = [(0, 1), (1, 9), (9, 15), (15, 36), (36, 42)]
                for key, ex in zip(keys, expected):
                    got = mapper.key2range(key)
                    self.assertEqual(got, ex)

                try:
                    mapper.key2range('bad')
                except KeyError:
                    pass
                else:
                    self.fail("Expected KeyError")

    def test_key2size(self):
        for mclass in (RangeTree, FlatRangeMapper):
            with self.subTest(msg=f'{mclass.__name__} test'):
                mapper = mclass(_data.items())
                keys = ['a', 'b', 'x', 'y', 'z']
                expected = [1, 8, 6, 21, 6]
                for key, ex in zip(keys, expected):
                    got = mapper.key2size(key)
                    self.assertEqual(got, ex)

                try:
                    mapper.key2size('bad')
                except KeyError:
                    pass
                else:
                    self.fail("Expected KeyError")

    def test_index2key_rel(self):
        for mclass in (RangeTree, FlatRangeMapper):
            with self.subTest(msg=f'{mclass.__name__} test'):
                mapper = mclass(_data.items())
                inds = [0, 1, 7, 9, 14, 15, 22, 41, 42, 43]
                expected = [('a',0), ('b',0), ('b',6), ('x',0), ('x',5), ('y',0), ('y',7), ('z',5), (None, None), (None, None)]
                for i, ex in zip(inds, expected):
                    got = mapper.index2key_rel(i)
                    self.assertEqual(got, ex)

    def test_iter(self):
        for mclass in (RangeTree, FlatRangeMapper):
            with self.subTest(msg=f'{mclass.__name__} test'):
                mapper = mclass(_data.items())
                expected = [('a', 0, 1), ('b', 1, 9), ('x', 9, 15), ('y', 15, 36), ('z', 36, 42)]
                for got, ex in zip(mapper, expected):
                    self.assertEqual(got, ex)

if __name__ == "__main__":
    unittest.main()



"""
Base class for interpolation methods that work on a semi-structured grid.
"""
import numpy as np

from openmdao.components.interp_util.interp_akima import InterpAkimaSemi
from openmdao.components.interp_util.interp_lagrange2 import InterpLagrange2Semi
from openmdao.components.interp_util.interp_lagrange3 import InterpLagrange3Semi
from openmdao.components.interp_util.interp_slinear import InterpLinearSemi
from openmdao.components.interp_util.outofbounds_error import OutOfBoundsError


INTERP_METHODS = {
    'slinear': InterpLinearSemi,
    'lagrange2': InterpLagrange2Semi,
    'lagrange3': InterpLagrange3Semi,
    'akima': InterpAkimaSemi,
}

TABLE_METHODS = [method for method in INTERP_METHODS.keys()]


class InterpNDSemi(object):
    """
    Interpolation on a semi-structured grid of arbitrary dimensions.

    Parameters
    ----------
    points : ndarray
        The points defining the semi-structured grid in n dimensions.
    values : ndarray
        Values (or initial values) for all points in the semi-structured grid.
    method : str
        Name of interpolation method.
    extrapolate : bool
        When False, raise an exception for any point that is extrapolated.
        When True, raise a warning for any point that is extrapolated.
        Note, default is True because semi-structured grids are often sparse in at least one
        dimension.
    **kwargs : dict
        Interpolator-specific options to pass onward.

    Attributes
    ----------
    extrapolate : bool
        When False, raise an exception for any point that is extrapolated.
        When True, raise a warning for any point that is extrapolated.
        Default is True (just warn).
    grid : tuple
        Collection of points that determine the regular grid.
    table : <InterpTable>
        Table object that contains algorithm that performs the interpolation.
    values : array_like, shape (m1, ..., mn, ...)
        The data on the regular grid in n dimensions.
    _compute_d_dvalues : bool
        When set to True, compute gradients with respect to the grid values.
    _d_dx : ndarray
        Cache of computed gradients with respect to evaluation point.
    _d_dvalues : ndarray
        Cache of computed gradients with respect to table values.
    _interp : class
        Class specified as interpolation algorithm, used to regenerate if needed.
    _interp_options : dict
        Dictionary of cached interpolator-specific options.
    _xi : ndarray
        Cache of current evaluation point.
    """

    def __init__(self, points, values, method="slinear", extrapolate=True, **kwargs):
        """
        Initialize an InterpNDSemi object.

        This object can be setup and used to interpolate on a curve or multi-dimensional table.
        """
        if not isinstance(method, str):
            msg = "Argument 'method' should be a string."
            raise ValueError(msg)
        elif method not in INTERP_METHODS:
            all_m = ', '.join(['"' + m + '"' for m in INTERP_METHODS])
            raise ValueError('Interpolation method "%s" is not defined. Valid methods are '
                             '%s.' % (method, all_m))

        if hasattr(values, 'dtype') and hasattr(values, 'astype'):
            if not np.issubdtype(values.dtype, np.inexact):
                values = values.astype(float)

        if len(points) > len(values):
            raise ValueError(f"There are {len(points)} point arrays, but {len(values)} values.")

        if np.iscomplexobj(values[:]):
            msg = "Interpolation method '%s' does not support complex values." % method
            raise ValueError(msg)

        self.grid = points
        self.values = values
        self.extrapolate = extrapolate

        self._xi = None
        self._d_dx = None
        self._d_dvalues = None
        self._compute_d_dvalues = False

        # Cache spline coefficients.
        interp = INTERP_METHODS[method]

        table = interp(self.grid, values, interp, **kwargs)
        table.check_config()

        self.table = table
        self._interp = interp
        self._interp_options = kwargs

    def interpolate(self, x, compute_derivative=False):
        """
        Interpolate at the sample coordinates.

        Parameters
        ----------
        x : ndarray or tuple
            Locations to interpolate.
        compute_derivative : bool
            Set to True to compute derivatives with respect to x.

        Returns
        -------
        ndarray
            Value of interpolant at all sample points.
        ndarray
            Value of derivative of interpolated output with respect to input x. (Only when
            compute_derivative is True).
        """
        xnew = self._interpolate(x)

        if compute_derivative:
            return xnew, self._d_dx
        else:
            return xnew

    def _interpolate(self, xi):
        """
        Interpolate at the sample coordinates.

        This method is called from OpenMDAO, and is not meant for standalone use.

        Parameters
        ----------
        xi : ndarray of shape (..., ndim)
            The coordinates to sample the gridded data.

        Returns
        -------
        ndarray
            Value of interpolant at all sample points.
        """
        # cache latest evaluation point for gradient method's use later
        self._xi = xi

        if not self.extrapolate:
            for i, p in enumerate(xi.T):
                if np.isnan(p).any():
                    raise OutOfBoundsError("One of the requested xi contains a NaN",
                                           i, np.NaN, self.grid[i][0], self.grid[i][-1])

        if self._compute_d_dvalues:
            # If the table grid or values are component inputs, then we need to create a new table
            # each iteration.
            interp = self._interp
            self.table = interp(self.grid, self.values, interp, extrapolate=self.extrapolate,
                                compute_d_dvalues=True, **self._interp_options)

        table = self.table

        xi = np.atleast_2d(xi)
        n_nodes, nx = xi.shape
        result = np.empty((n_nodes, ), dtype=xi.dtype)
        derivs_x = np.empty((n_nodes, nx), dtype=xi.dtype)
        if self._compute_d_dvalues:
            derivs_val = np.zeros((n_nodes, len(self.values)), dtype=xi.dtype)

        # Loop over n_nodes because there isn't a way to vectorize.
        for j in range(n_nodes):
            val, d_x, d_values_tuple, extrapolate = table.interpolate(xi[j, :])
            result[j] = val.item()
            derivs_x[j, :] = d_x.ravel()
            if self._compute_d_dvalues:
                d_values, idx = d_values_tuple
                derivs_val[j, idx] = d_values.ravel()

        # Cache derivatives
        self._d_dx = derivs_x
        if self._compute_d_dvalues:
            self._d_dvalues = derivs_val

        return result

    def gradient(self, xi):
        """
        Compute the gradients at the specified point.

        Most of the gradients are computed as the interpolation itself is performed,
        but are cached and returned separately by this method.

        If the point for evaluation differs from the point used to produce
        the currently cached gradient, the interpolation is re-performed in
        order to return the correct gradient.

        Parameters
        ----------
        xi : ndarray of shape (..., ndim)
            The coordinates to sample the gridded data at.

        Returns
        -------
        ndarray
            Vector of gradients of the interpolated values with respect to each value in xi.
        """
        if (self._xi is None) or (not np.array_equal(xi, self._xi)):
            # If inputs have changed since last computation, then re-interpolate.
            self.interpolate(xi)

        return self._d_dx

    def training_gradients(self):
        """
        Compute the training gradient for the vector of training points.

        Returns
        -------
        ndarray
            Gradient of output with respect to training point values.
        """
        return self._d_dvalues



import unittest

import numpy as np

import openmdao.api as om
import openmdao.utils.coloring as coloring_mod
from openmdao.utils.assert_utils import assert_near_equal, assert_check_totals
from openmdao.utils.general_utils import set_pyoptsparse_opt
from openmdao.utils.testing_utils import use_tempdirs


# check that pyoptsparse is installed
OPT, OPTIMIZER = set_pyoptsparse_opt('SLSQP')

if OPTIMIZER:
    from openmdao.drivers.pyoptsparse_driver import pyOptSparseDriver


class TrajDesignParameterOptionsDictionary(om.OptionsDictionary):

    def __init__(self, read_only=False):
        super().__init__(read_only)

        self.declare(name='name')
        self.declare(name='val', default=np.zeros(1))
        self.declare(name='targets', types=dict, default=None, allow_none=True)
        self.declare(name='shape', default=(1,))


class StateOptionsDictionary(om.OptionsDictionary):

    def __init__(self, read_only=False):
        super().__init__(read_only)

        self.declare(name='name')
        self.declare(name='val', default=0.0)
        self.declare(name='shape', default=(1,))
        self.declare(name='rate_source')


class CollocationComp(om.ExplicitComponent):

    def initialize(self):

        self.options.declare('state_options', types=dict)

    def setup(self):
        num_col_nodes = 1
        state_options = self.options['state_options']

        self.var_names = var_names = {}
        for state_name in state_options:
            var_names[state_name] = {
                'f_approx': 'f_approx:{0}'.format(state_name),
                'f_computed': 'f_computed:{0}'.format(state_name),
                'defect': 'defects:{0}'.format(state_name),
            }

        for state_name, options in state_options.items():
            shape = options['shape']

            var_names = self.var_names[state_name]

            self.add_input(
                name=var_names['f_approx'],
                shape=(num_col_nodes,) + shape)

            self.add_input(
                name=var_names['f_computed'],
                shape=(num_col_nodes,) + shape)

            self.add_output(
                name=var_names['defect'],
                shape=(num_col_nodes,) + shape)

            self.add_constraint(name=var_names['defect'], equals=0.0)

        # Setup partials
        state_options = self.options['state_options']

        for state_name, options in state_options.items():
            shape = options['shape']
            size = np.prod(shape)

            r = np.arange(num_col_nodes * size)

            var_names = self.var_names[state_name]

            self.declare_partials(of=var_names['defect'],
                                  wrt=var_names['f_approx'],
                                  rows=r, cols=r)

            self.declare_partials(of=var_names['defect'],
                                  wrt=var_names['f_computed'],
                                  rows=r, cols=r)

    def compute(self, inputs, outputs):
        state_options = self.options['state_options']

        for state_name in state_options:
            var_names = self.var_names[state_name]

            f_approx = inputs[var_names['f_approx']]
            f_computed = inputs[var_names['f_computed']]

            outputs[var_names['defect']] = ((f_approx - f_computed).T).T

    def compute_partials(self, inputs, partials):
        for state_name, options in self.options['state_options'].items():
            size = np.prod(options['shape'])
            var_names = self.var_names[state_name]

            k = np.repeat(1, size)

            partials[var_names['defect'], var_names['f_approx']] = k
            partials[var_names['defect'], var_names['f_computed']] = -k


class StateInterpComp(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('state_options', types=dict)

    def setup(self):
        num_disc_nodes = 2
        num_col_nodes = 1

        state_options = self.options['state_options']

        self.xd_str = {}
        self.fd_str = {}
        self.xc_str = {}
        self.xdotc_str = {}

        for state_name, options in state_options.items():
            shape = options['shape']

            self.add_input(
                name='state_disc:{0}'.format(state_name),
                shape=(num_disc_nodes,) + shape)

            self.add_input(
                name='staterate_disc:{0}'.format(state_name),
                shape=(num_disc_nodes,) + shape)

            self.add_output(
                name='state_col:{0}'.format(state_name),
                shape=(num_col_nodes,) + shape)

            self.add_output(
                name='staterate_col:{0}'.format(state_name),
                shape=(num_col_nodes,) + shape)

            self.xd_str[state_name] = 'state_disc:{0}'.format(state_name)
            self.fd_str[state_name] = 'staterate_disc:{0}'.format(state_name)
            self.xc_str[state_name] = 'state_col:{0}'.format(state_name)
            self.xdotc_str[state_name] = 'staterate_col:{0}'.format(state_name)

        Ai = np.array([[0.5, 0.5]])
        Bi = np.array([[ 0.25, -0.25]])
        Ad = np.array([[-0.75,  0.75]])
        Bd = np.array([[-0.25, -0.25]])

        self.matrices = {'Ai': Ai, 'Bi': Bi, 'Ad': Ad, 'Bd': Bd}

        # Setup partials

        self.jacs = {'Ai': {}, 'Bi': {}, 'Ad': {}, 'Bd': {}}
        self.Bi_rows = {}
        self.Bi_cols = {}
        self.Ad_rows = {}
        self.Ad_cols = {}
        self.sizes = {}
        self.num_disc_nodes = num_disc_nodes
        for name, options in state_options.items():
            shape = options['shape']

            size = np.prod(shape)

            for key in self.jacs:
                jac = np.zeros((num_col_nodes, size, num_disc_nodes, size))
                for i in range(size):
                    jac[:, i, :, i] = self.matrices[key]
                jac = jac.reshape((num_col_nodes * size, num_disc_nodes * size), order='C')
                self.jacs[key][name] = jac

            self.sizes[name] = size

            rs_dtdstau = np.zeros(num_col_nodes * size, dtype=int)
            r_band = np.arange(0, num_col_nodes, dtype=int) * size

            r0 = 0
            for i in range(size):
                rs_dtdstau[r0:r0 + num_col_nodes] = r_band + i
                r0 += num_col_nodes

            Ai_rows, Ai_cols = np.where(self.jacs['Ai'][name] != 0)
            self.declare_partials(of=self.xc_str[name], wrt=self.xd_str[name],
                                  rows=Ai_rows, cols=Ai_cols,
                                  val=self.jacs['Ai'][name][Ai_rows, Ai_cols])

            self.Bi_rows[name], self.Bi_cols[name] = np.where(self.jacs['Bi'][name] != 0)
            self.declare_partials(of=self.xc_str[name], wrt=self.fd_str[name],
                                  rows=self.Bi_rows[name], cols=self.Bi_cols[name])

            Bd_rows, Bd_cols = np.where(self.jacs['Bd'][name] != 0)
            self.declare_partials(of=self.xdotc_str[name], wrt=self.fd_str[name],
                                  rows=Bd_rows, cols=Bd_cols,
                                  val=self.jacs['Bd'][name][Bd_rows, Bd_cols])

            self.Ad_rows[name], self.Ad_cols[name] = np.where(self.jacs['Ad'][name] != 0)
            self.declare_partials(of=self.xdotc_str[name], wrt=self.xd_str[name],
                                  rows=self.Ad_rows[name], cols=self.Ad_cols[name])

    def compute(self, inputs, outputs):
        state_options = self.options['state_options']

        for name in state_options:
            xc_str = self.xc_str[name]
            xdotc_str = self.xdotc_str[name]
            xd_str = self.xd_str[name]
            fd_str = self.fd_str[name]

            xd = np.atleast_2d(inputs[xd_str])

            a = np.tensordot(self.matrices['Bi'], inputs[fd_str], axes=(1, 0)).T
            outputs[xc_str] = a.T

            outputs[xc_str] += np.tensordot(
                self.matrices['Ai'], xd, axes=(1, 0))

            outputs[xdotc_str] = np.tensordot(self.matrices['Ad'], xd, axes=(1, 0))

            outputs[xdotc_str] += np.tensordot(
                self.matrices['Bd'], inputs[fd_str], axes=(1, 0))

    def compute_partials(self, inputs, partials):
        for name, options in self.options['state_options'].items():
            xdotc_name = self.xdotc_str[name]
            xd_name = self.xd_str[name]

            xc_name = self.xc_str[name]
            fd_name = self.fd_str[name]

            r_nz, c_nz = self.Bi_rows[name], self.Bi_cols[name]
            partials[xc_name, fd_name] = (self.jacs['Bi'][name])[r_nz, c_nz]

            r_nz, c_nz = self.Ad_rows[name], self.Ad_cols[name]
            partials[xdotc_name, xd_name] = (self.jacs['Ad'][name])[r_nz, c_nz]


class StateIndependentsComp(om.ImplicitComponent):

    def initialize(self):

        self.options.declare('state_options', types=dict)

    def setup(self):
        state_options = self.options['state_options']

        num_state_input_nodes = 2

        var_names = {}
        for state_name in state_options:
            var_names[state_name] = {
                'defect': 'defects:{0}'.format(state_name),
            }

        for state_name, options in state_options.items():

            shape = options['shape']
            var_names = var_names[state_name]

            self.add_output(name='states:{0}'.format(state_name),
                            shape=(num_state_input_nodes, ) + shape)

        # Setup partials
        for state_name, options in state_options.items():
            shape = options['shape']
            state_var_name = 'states:{0}'.format(state_name)

            row_col = np.arange(num_state_input_nodes*np.prod(shape))
            self.declare_partials(of=state_var_name, wrt=state_var_name,
                                  rows=row_col, cols=row_col, val=-1.0)

class Trajectory(om.Group):

    def __init__(self):
        super().__init__()

        self.design_parameter_options = {}
        self._phases = {}

    def add_phase(self, name, phase):
        self._phases[name] = phase
        return phase

    def add_design_parameter(self, name, val, targets):
        if name not in self.design_parameter_options:
            self.design_parameter_options[name] = TrajDesignParameterOptionsDictionary()

        self.design_parameter_options[name]['val'] = val
        self.design_parameter_options[name]['targets'] = targets


    def _setup_design_parameters(self):
        if self.design_parameter_options:
            indep = self.add_subsystem('design_params', subsys=om.IndepVarComp(),
                                       promotes_outputs=['*'])

            for name, options in self.design_parameter_options.items():
                indep.add_output(name='design_parameters:{0}'.format(name),
                                 val=options['val'],
                                 shape=(1, np.prod(options['shape'])))

    def setup(self):
        super().setup()
        self._setup_design_parameters()

        phases_group = self.add_subsystem('phases', subsys=om.ParallelGroup(), promotes_inputs=['*'],
                                          promotes_outputs=['*'])

        for name, phs in self._phases.items():
            g = phases_group.add_subsystem(name, phs)
            g.linear_solver = om.DirectSolver()


class Phase(om.Group):

    def __init__(self, **kwargs):

        _kwargs = kwargs.copy()

        self.state_options = {}
        self._objectives = {}

        super().__init__(**_kwargs)

    def initialize(self):
        self.options.declare('ode_class', default=None)
        self.options.declare('transcription')

    def add_state(self, name, rate_source):
        if name not in self.state_options:
            self.state_options[name] = StateOptionsDictionary()
            self.state_options[name]['name'] = name

        self.set_state_options(name=name, rate_source=rate_source)

    def set_state_options(self, name, rate_source):
        self.state_options[name]['rate_source'] = rate_source

    def add_objective(self, name, loc='final', index=None, shape=(1,)):
        obj_dict = {'loc': loc,
                    'index': index,
                    'shape': shape}
        self._objectives[name] = obj_dict

    def setup(self):
        transcription = self.options['transcription']

        transcription.setup_states(self)
        transcription.setup_ode(self)
        transcription.setup_defects(self)

        transcription.setup_objective(self)


class GaussLobatto(object):

    def setup_objective(self, phase):
        for name, options in phase._objectives.items():
            index = options['index']

            shape = phase.state_options[name]['shape']
            obj_path = 'states:{0}'.format(name)

            shape = options['shape'] if shape is None else shape

            size = int(np.prod(shape))
            idx = 0 if index is None else index
            obj_index = -size + idx

            super(Phase, phase).add_objective(obj_path, index=obj_index, flat_indices=True)

    def setup_states(self, phase):
        indep = StateIndependentsComp(state_options=phase.state_options)

        num_connected = 0
        prom_inputs = ['initial_states:*'] if num_connected > 0 else None
        phase.add_subsystem('indep_states', indep, promotes_inputs=prom_inputs,
                            promotes_outputs=['*'])

        for name, options in phase.state_options.items():
            size = np.prod(options['shape'])
            desvar_indices = list(range(size * 2))

            phase.add_design_var(name='states:{0}'.format(name),
                                 indices=desvar_indices, flat_indices=True)

    def setup_ode(self, phase):
        map_input_indices_to_disc = np.array([0, 1])

        phase.add_subsystem('state_interp',
                            subsys=StateInterpComp(state_options=phase.state_options))

        for name, options in phase.state_options.items():
            size = np.prod(options['shape'])

            src_idxs_mat = np.reshape(np.arange(size * 2, dtype=int),
                                      (2, size), order='C')

            src_idxs = src_idxs_mat[map_input_indices_to_disc, :].flat[:]

            phase.connect('states:{0}'.format(name),
                          'state_interp.state_disc:{0}'.format(name),
                          src_indices=src_idxs, flat_src_indices=True)

    def setup_defects(self, phase):

        phase.add_subsystem('collocation_constraint',
                            CollocationComp(state_options=phase.state_options))

        for name in phase.state_options:
            phase.connect('state_interp.staterate_col:{0}'.format(name),
                          'collocation_constraint.f_approx:{0}'.format(name))


class FiniteBurnODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('accel', val=np.zeros(nn))
        self.add_input('c', val=np.zeros(nn))

        self.add_output('deltav_dot', val=np.zeros(nn))

        ar = np.arange(self.options['num_nodes'])
        self.declare_partials(of='deltav_dot', wrt='accel', rows=ar, cols=ar, val=1.0)

    def compute(self, inputs, outputs):
        at = inputs['accel']
        outputs['deltav_dot'] = at


def make_traj():

    t = GaussLobatto()

    traj = Trajectory()

    traj.add_design_parameter('c', val=1.5, targets={'burn1': ['c'], 'burn2': ['c']})

    # First Phase (burn)
    burn1 = Phase(ode_class=FiniteBurnODE, transcription=t)
    burn1 = traj.add_phase('burn1', burn1)
    burn1.add_state('deltav', rate_source='deltav_dot')

    # Third Phase (burn)
    burn2 = Phase(ode_class=FiniteBurnODE, transcription=t)
    traj.add_phase('burn2', burn2)
    burn2.add_state('deltav', rate_source='deltav_dot')

    burn2.add_objective('deltav', loc='final')

    return traj


@use_tempdirs
@unittest.skipUnless(OPTIMIZER, "This test requires pyOptSparseDriver.")
class TestMPIColoringBug(unittest.TestCase):
    N_PROCS = 2

    def test_bug(self):
        class ColoringOnly(om.pyOptSparseDriver):

            def run(self):
                """
                Color the model.
                """
                if coloring_mod._use_total_sparsity:
                    if self._coloring_info.do_compute_coloring() and self._coloring_info['dynamic']:
                        coloring_mod.dynamic_total_coloring(self, run_model=True,
                                                            fname=self._get_total_coloring_fname())
                        self._setup_tot_jac_sparsity()

        p = om.Problem()

        p.driver = ColoringOnly()
        p.driver.declare_coloring()

        p.model = make_traj()

        p.setup(mode='rev')

        # Set Initial Guesses
        p.set_val('design_parameters:c', val=1.5)

        of = ['phases.burn2.indep_states.states:deltav', 'phases.burn1.collocation_constraint.defects:deltav', 'phases.burn2.collocation_constraint.defects:deltav', ]
        wrt = ['phases.burn1.indep_states.states:deltav', 'phases.burn2.indep_states.states:deltav']

        p.run_model()
        p.run_driver()

        J = p.driver._compute_totals(of=of, wrt=wrt, return_format='dict')
        dd = J['phases.burn1.collocation_constraint.defects:deltav']['phases.burn1.indep_states.states:deltav']

        assert_near_equal(dd, np.array([[-0.75, 0.75]]), 1e-6)

    def test_bug2(self):
        size = 3
        p = om.Problem()
        phases = p.model.add_subsystem('phases', om.ParallelGroup())
        phase1 = phases.add_subsystem('phase1', om.Group())
        phase1.add_subsystem('indep', om.IndepVarComp('x', val=np.ones(size)))
        phase1.add_subsystem('comp1', om.ExecComp('y=3.0*x', x=np.ones(size), y=np.ones(size)))
        phase1.add_subsystem('comp2', om.ExecComp('y=5.0*x', x=np.ones(size), y=np.ones(size)))
        phase1.connect('indep.x', 'comp1.x')
        phase1.connect('comp1.y', 'comp2.x')
        phase1.add_design_var('indep.x')
        phase1.add_constraint('comp2.y', lower=0.0)
        phase1.add_constraint('comp1.y', lower=0.0)

        phase2 = phases.add_subsystem('phase2', om.Group())
        phase2.add_subsystem('indep', om.IndepVarComp('x', val=np.ones(size)))
        phase2.add_subsystem('comp1', om.ExecComp('y=7.0*x', x=np.ones(size), y=np.ones(size)))
        phase2.add_subsystem('comp2', om.ExecComp('y=9.0*x', x=np.ones(size), y=np.ones(size)))
        phase2.connect('indep.x', 'comp1.x')
        phase2.connect('comp1.y', 'comp2.x')
        phase2.add_design_var('indep.x')
        phase2.add_constraint('comp2.y', lower=0.0)
        phase2.add_constraint('comp1.y', lower=0.0)

        p.setup(mode='rev')
        p.run_model()

        assert_check_totals(p.check_totals())



"""
This allows you to run a specific test function by itself from an importable module.

The function can be either a method of a unittest.TestCase subclass, or just a function defined
at module level.  This is useful when running an individual test under mpirun for debugging
purposes.

To specify the test to run, use the following forms:

<modpath>:<testcase name>.<funcname>   OR   <modpath>:<funcname>

where <modpath> is either the dotted module name or the full filesystem path of the python file.

for example:

    mpirun -n 4 run_test mypackage.mysubpackage.mymod:MyTestCase.test_foo

    OR

    mpirun -n 4 run_test /foo/bar/mypackage/mypackage/mysubpackage/mymod.py:MyTestCase.test_foo
"""

import os
import sys
import importlib
from openmdao.utils.file_utils import get_module_path


def run_test():
    """
    Run individual test(s).
    """
    if len(sys.argv) > 1:
        testspec = sys.argv[1]
        parts = testspec.split(':')

    if len(sys.argv) != 2 or len(parts) != 2:
        print('Usage: run_test my_mod_path:my_test_case.test_func_name\n'
              '            OR\n'
              '       run_test my_mod_path:test_func_name')
        sys.exit(-1)

    modpath, funcpath = parts
    if modpath.endswith('.py'):
        modpath = get_module_path(modpath)
        if modpath is None:
            modpath = parts[0]
            moddir = os.path.dirname(modpath)
            sys.path = [moddir] + sys.path
            modpath = os.path.basename(modpath)[:-3]

    mod = importlib.import_module(modpath)

    parts = funcpath.split('.', 1)
    if len(parts) == 2:
        tcase_name, method_name = parts
        testcase = getattr(mod, tcase_name)(methodName=method_name)
        setup = getattr(testcase, 'setUp', None)
        if setup is not None:
            setup()
        getattr(testcase, method_name)()
        teardown = getattr(testcase, 'tearDown', None)
        if teardown:
            teardown()
    else:
        funcname = parts[0]
        getattr(mod, funcname)()


if __name__ == '__main__':
    run_test()



"""Define the explicit test component (array)."""
import numpy as np
import scipy.sparse

import openmdao.api as om


class TestExplCompArray(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('thickness', default=1.)

    def setup(self):
        self.add_input('lengths', val=np.ones((2, 2)))
        self.add_input('widths', val=np.ones((2, 2)))
        self.add_output('areas', val=np.ones((2, 2)))
        self.add_output('total_volume', val=1.)

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute(self, inputs, outputs):
        thk = self.options['thickness']

        outputs['areas'] = inputs['lengths'] * inputs['widths']
        outputs['total_volume'] = np.sum(outputs['areas']) * thk


class TestExplCompArrayDense(TestExplCompArray):

    def compute_partials(self, inputs, partials):
        thk = self.options['thickness']

        partials['areas', 'lengths'] = np.diag(inputs['widths'].flatten())
        partials['areas', 'widths'] = np.diag(inputs['lengths'].flatten())
        partials['total_volume', 'lengths'] = inputs['widths'].flatten() * thk
        partials['total_volume', 'widths'] = inputs['lengths'].flatten() * thk


class TestExplCompArraySparse(TestExplCompArray):

    def setup_partials(self):
        self.declare_partials('areas', ['lengths', 'widths'], rows=np.arange(4), cols=np.arange(4))
        self.declare_partials('total_volume', ['lengths', 'widths'])

    def compute_partials(self, inputs, partials):
        thk = self.options['thickness']

        partials['areas', 'lengths'] = inputs['widths'].flatten()
        partials['areas', 'widths'] = inputs['lengths'].flatten()
        partials['total_volume', 'lengths'] = inputs['widths'].flatten() * thk
        partials['total_volume', 'widths'] = inputs['lengths'].flatten() * thk


class TestExplCompArrayJacVec(TestExplCompArray):

    def setup_partials(self):
        pass  # prevent declared partials from base class

    def compute_jacvec_product(self, inputs, d_inputs, d_outputs, mode):

        if mode == 'fwd':
            if 'areas' in d_outputs:
                if 'widths' in d_inputs:
                    d_outputs['areas'] += d_inputs['widths']*inputs['lengths']
                if 'lengths' in d_inputs:
                    d_outputs['areas'] += d_inputs['lengths']*inputs['widths']
            if 'total_volume' in d_outputs:
                if 'widths' in d_inputs:
                    d_outputs['total_volume'] += np.sum(d_inputs['widths']*inputs['lengths'])
                if 'lengths' in d_inputs:
                    d_outputs['total_volume'] += np.sum(d_inputs['lengths']*inputs['widths'])
        else:
            if 'widths' in d_inputs:
                d_inputs['widths'] += d_outputs['areas']*inputs['lengths']
                d_inputs['widths'] += d_outputs['total_volume']*inputs['lengths']
            if 'lengths' in d_inputs:
                d_inputs['lengths'] += d_outputs['areas']*inputs['widths']
                d_inputs['lengths'] += d_outputs['total_volume']*inputs['widths']



import unittest
import numpy as np

import openmdao.api as om
from openmdao.utils.mpi import MPI
from openmdao.utils.assert_utils import assert_near_equal, assert_check_partials, \
     assert_check_totals
from openmdao.test_suite.groups.parallel_groups import FanIn, FanOut

try:
    from openmdao.vectors.petsc_vector import PETScVector
except ImportError:
    PETScVector = None


def build_submodelcomp1(promote=True, **kwargs):
    subprob1 = om.Problem()
    submodel1 = subprob1.model.add_subsystem('submodel1', om.Group())
    submodel1.add_subsystem('sub1_ivc_r', om.IndepVarComp('r', 1.),
                            promotes_outputs=['r'])
    submodel1.add_subsystem('sub1_ivc_theta', om.IndepVarComp('theta', np.pi),
                            promotes_outputs=['theta'])
    if promote:
        promotes = ['*']
    else:
        promotes = ()
    submodel1.add_subsystem('subComp1', om.ExecComp('x = r*cos(theta)'), promotes=promotes)
    subprob1.model.promotes('submodel1', any=['*'])
    return om.SubmodelComp(problem=subprob1, **kwargs)


def build_submodelcomp2(promote=True, **kwargs):
    subprob2 = om.Problem()
    submodel2 = subprob2.model.add_subsystem('submodel2', om.Group())
    submodel2.add_subsystem('sub2_ivc_r', om.IndepVarComp('r', 2),
                            promotes_outputs=['r'])
    submodel2.add_subsystem('sub2_ivc_theta', om.IndepVarComp('theta', np.pi/2),
                            promotes_outputs=['theta'])
    if promote:
        promotes = ['*']
    else:
        promotes = ()
    submodel2.add_subsystem('subComp2', om.ExecComp('y = r*sin(theta)'), promotes=promotes)
    subprob2.model.promotes('submodel2', any=['*'])
    return om.SubmodelComp(problem=subprob2, **kwargs)


class TestSubmodelComp(unittest.TestCase):
    def test_submodel_comp(self):
        p = om.Problem()
        supmodel = om.Group()
        supmodel.add_subsystem('supComp', om.ExecComp('z = x**2 + y'),
                               promotes_inputs=['x', 'y'],
                               promotes_outputs=['z'])

        p.model.add_subsystem('sub1', build_submodelcomp1(inputs=['r', 'theta'], outputs=['x']),
                              promotes_inputs=['r','theta'],
                              promotes_outputs=['x'])

        p.model.add_subsystem('sub2', build_submodelcomp2(inputs=['r', 'theta'], outputs=['y']),
                              promotes_inputs=['r','theta'],
                              promotes_outputs=['y'])

        p.model.add_subsystem('supModel', supmodel, promotes_inputs=['x','y'],
                              promotes_outputs=['z'])

        p.model.set_input_defaults('r', 1)
        p.model.set_input_defaults('theta', np.pi)

        p.setup(force_alloc_complex=True)

        p.run_model()
        assert_check_partials(p.check_partials(method='cs', out_stream=None))

        assert_near_equal(p.get_val('z'), 1.0)

    def test_no_io(self):
        p = om.Problem()

        model = om.Group()
        model.add_subsystem('supComp', om.ExecComp('z = x**2 + y'),
                            promotes_inputs=['x', 'y'],
                            promotes_outputs=['z'])

        p.model.add_subsystem('sub1', build_submodelcomp1())
        p.model.add_subsystem('sub2', build_submodelcomp2())
        p.model.add_subsystem('supModel', model, promotes_inputs=['x','y'], promotes_outputs=['z'])

        p.setup(force_alloc_complex=True)

        with self.assertRaises(Exception) as ctx:
            p.set_val('r', 1)

        msg = '\'<model> <class Group>: Variable "r" not found.\''
        self.assertTrue(str(ctx.exception).startswith(msg))

    def test_variable_alias(self):
        p = om.Problem()
        model = om.Group()

        model.add_subsystem('subsys', om.ExecComp('z = x**2 + y**2'))
        subprob = om.Problem()
        subprob.model.add_subsystem('submodel', model)
        subcomp = om.SubmodelComp(problem=subprob,
                                  inputs=[('submodel.subsys.x', 'a'), ('submodel.subsys.y', 'b')],
                                  outputs=[('submodel.subsys.z', 'c')])

        p.model.add_subsystem('subcomp', subcomp, promotes_inputs=['a', 'b'], promotes_outputs=['c'])
        p.setup()

        p.set_val('a', 1)
        p.set_val('b', 2)

        p.run_model()

        inputs = p.model.subcomp.list_inputs()
        outputs = p.model.subcomp.list_outputs()

        inputs = {inputs[i][0]: inputs[i][1]['val'] for i in range(len(inputs))}
        outputs = {outputs[i][0]: outputs[i][1]['val'] for i in range(len(outputs))}

        assert_near_equal(inputs['a'], 1)
        assert_near_equal(inputs['b'], 2)
        assert_near_equal(outputs['c'], 5)

    def test_unconnected_same_var(self):
        p = om.Problem()

        model = om.Group()

        model.add_subsystem('x1Comp', om.ExecComp('x1 = x*3'))
        model.add_subsystem('x2Comp', om.ExecComp('x2 = x**3'))
        model.connect('x1Comp.x1', 'model.x1')
        model.connect('x2Comp.x2', 'model.x2')
        model.add_subsystem('model', om.ExecComp('z = x1**2 + x2**2'))

        subprob = om.Problem()
        subprob.model.add_subsystem('submodel', model)

        comp = om.SubmodelComp(problem=subprob,
                               inputs=[('submodel.x1Comp.x', 'x'), ('submodel.x2Comp.x', 'y')],
                               outputs=[('submodel.model.z', 'z')])

        p.model.add_subsystem('comp', comp)

        p.model.set_input_defaults('comp.x', 1)
        p.model.set_input_defaults('comp.y', 2)

        p.setup()

        p.run_model()

        inputs = p.model.comp.list_inputs()
        outputs = p.model.comp.list_outputs()

        inputs = {inputs[i][0]: inputs[i][1]['val'] for i in range(len(inputs))}
        outputs = {outputs[i][0]: outputs[i][1]['val'] for i in range(len(outputs))}

        assert_near_equal(inputs['x'], 1)
        assert_near_equal(inputs['y'], 2)
        assert_near_equal(outputs['z'], 73)

    def test_wildcard(self):
        p = om.Problem()
        model = om.Group()

        model.add_subsystem('sub', om.ExecComp('z = x1**2 + x2**2 + x3**2'), promotes=['*'])
        subprob = om.Problem()
        subprob.model.add_subsystem('submodel', model, promotes=['*'])
        comp = om.SubmodelComp(problem=subprob, inputs=['x*'], outputs=['*'])

        p.model.add_subsystem('comp', comp, promotes_inputs=['*'], promotes_outputs=['*'])
        p.setup()

        p.set_val('x1', 1)
        p.set_val('x2', 2)
        p.set_val('x3', 3)

        p.run_model()

        inputs = p.model.comp.list_inputs()
        outputs = p.model.comp.list_outputs()

        inputs = {inputs[i][0]: inputs[i][1]['val'] for i in range(len(inputs))}
        outputs = {outputs[i][0]: outputs[i][1]['val'] for i in range(len(outputs))}

        assert_near_equal(inputs['x1'], 1)
        assert_near_equal(inputs['x2'], 2)
        assert_near_equal(inputs['x3'], 3)
        assert_near_equal(outputs['z'], 14)

    def test_add_io_before_setup(self):
        p = om.Problem()

        model = om.Group()
        model.add_subsystem('supComp', om.ExecComp('z = x**2 + y'),
                            promotes_inputs=['x', 'y'],
                            promotes_outputs=['z'])

        comp1 = build_submodelcomp1(promote=False)
        comp2 = build_submodelcomp2(promote=False)

        comp1.add_input('subComp1.r', name='r')
        comp1.add_input('subComp1.theta', name='theta')
        comp2.add_input('subComp2.r', name='r')
        comp2.add_input('subComp2.theta', name='theta')

        comp1.add_output('subComp1.x', name='x')
        comp2.add_output('subComp2.y', name='y')

        p.model.add_subsystem('sub1', comp1, promotes_inputs=['r','theta'],
                                    promotes_outputs=['x'])
        p.model.add_subsystem('sub2', comp2, promotes_inputs=['r','theta'],
                                    promotes_outputs=['y'])
        p.model.add_subsystem('supModel', model, promotes_inputs=['x','y'],
                                    promotes_outputs=['z'])

        p.model.set_input_defaults('r', 1)
        p.model.set_input_defaults('theta', np.pi)

        p.setup(force_alloc_complex=True)

        p.run_model()
        cpd = p.check_partials(method='cs', out_stream=None)

        assert_near_equal(p.get_val('z'), 1.0)

    def test_add_io_after_setup(self):
        class Subsys1(om.Group):
            def setup(self):
                model = om.Group()
                comp = om.ExecComp('x = r*cos(theta)')
                model.add_subsystem('comp', comp, promotes_inputs=['r', 'theta'],
                                    promotes_outputs=['x'])
                subprob = om.Problem(); subprob.model.add_subsystem('model', model)
                subprob.model.promotes('model', any=['*'])
                self.add_subsystem('submodel1', om.SubmodelComp(problem=subprob))

            def configure(self):
                self._get_subsystem('submodel1').add_input('r')
                self._get_subsystem('submodel1').add_input('theta')
                self._get_subsystem('submodel1').add_output('x')

                self.promotes('submodel1', ['r', 'theta', 'x'])

        class Subsys2(om.Group):
            def setup(self):
                model = om.Group()
                comp = om.ExecComp('y = r*sin(theta)')
                model.add_subsystem('comp', comp, promotes_inputs=['r', 'theta'],
                                    promotes_outputs=['y'])
                subprob = om.Problem(); subprob.model.add_subsystem('model', model)
                subprob.model.promotes('model', any=['*'])
                self.add_subsystem('submodel2', om.SubmodelComp(problem=subprob))

            def configure(self):
                self._get_subsystem('submodel2').add_input('r')
                self._get_subsystem('submodel2').add_input('theta')
                self._get_subsystem('submodel2').add_output('y')

                self.promotes('submodel2', ['r', 'theta', 'y'])

        p = om.Problem()

        model = om.Group()
        model.add_subsystem('supComp', om.ExecComp('z = x**2 + y'),
                            promotes_inputs=['x', 'y'],
                            promotes_outputs=['z'])

        p.model.add_subsystem('sub1', Subsys1(), promotes_inputs=['r', 'theta'],
                              promotes_outputs=['x'])
        p.model.add_subsystem('sub2', Subsys2(), promotes_inputs=['r', 'theta'],
                              promotes_outputs=['y'])
        p.model.add_subsystem('supModel', model, promotes_inputs=['x','y'],
                              promotes_outputs=['z'])

        p.setup(force_alloc_complex=True)

        p.set_val('r', 1)
        p.set_val('theta', np.pi)

        p.run_model()
        cpd = p.check_partials(method='cs', out_stream=None)

        assert_near_equal(p.get_val('z'), 1)

    def test_invalid_io(self):
        p = om.Problem()

        model = om.Group()
        model.add_subsystem('supComp', om.ExecComp('z = x**2 + y'),
                            promotes_inputs=['x', 'y'],
                            promotes_outputs=['z'])

        comp1 = build_submodelcomp1()
        comp2 = build_submodelcomp2()

        comp1.add_input('psi')

        p.model.add_subsystem('sub1', comp1)
        p.model.add_subsystem('sub2', comp2)
        p.model.add_subsystem('supModel', model, promotes_inputs=['x','y'],
                              promotes_outputs=['z'])

        with self.assertRaises(Exception) as ctx:
            p.setup(force_alloc_complex=True)

        msg = "'psi' is not an independent variable in the submodel."
        self.assertEqual(str(ctx.exception), msg)

    def test_multiple_setups(self):
        p = om.Problem()

        model = om.Group()
        model.add_subsystem('supComp', om.ExecComp('y = 3*x + 4'),
                            promotes_inputs=['x'], promotes_outputs=['y'])

        submodel = om.Group()
        submodel.add_subsystem('subComp', om.ExecComp('x = 6*z + 3'),
                               promotes_inputs=['z'], promotes_outputs=['x'])

        subprob = om.Problem()
        subprob.model.add_subsystem('submodel', submodel, promotes=['*'])

        comp = om.SubmodelComp(problem=subprob)
        comp.add_input('z')
        comp.add_output('x')

        p.model.add_subsystem('sub', comp, promotes_inputs=['z'],
                              promotes_outputs=['x'])
        p.model.add_subsystem('sup', model, promotes_inputs=['x'],
                              promotes_outputs=['y'])

        p.setup(force_alloc_complex=False)
        p.setup(force_alloc_complex=False)
        p.setup(force_alloc_complex=True)

        p.set_val('z', 1)

        p.run_model()

        assert_near_equal(p.get_val('z'), 1)
        assert_near_equal(p.get_val('x'), 9)
        assert_near_equal(p.get_val('y'), 31)

    def test_multiple_submodel_setups(self):
        top = om.Problem()
        model = top.model
        model.add_subsystem('supComp', om.ExecComp('y = 3*x + 4'),
                            promotes_inputs=['x'], promotes_outputs=['y'])

        submodel = om.Group()
        submodel.add_subsystem('subComp', om.ExecComp('x = 6*z + 3'),
                               promotes_inputs=['z'], promotes_outputs=['x'])

        subprob = om.Problem()
        subprob.model.add_subsystem('submodel', submodel, promotes=['*'])

        comp = om.SubmodelComp(problem=subprob)
        comp.add_input('z')
        comp.add_output('x')

        top.setup()
        top.setup()
        top.final_setup()
        top.run_model()

    def test_add_io_meta_override(self):
        p = om.Problem()
        subprob = om.Problem()
        subprob.model.add_subsystem('comp', om.ExecComp('x = r*cos(theta)'), promotes=['*'])
        submodel = om.SubmodelComp(problem=subprob)

        submodel.add_input('r', name='new_r', val=20)
        submodel.add_input('theta', name='new_theta', val=0.5)
        submodel.add_output('x', name='new_x', val=100)

        p.model.add_subsystem('submodel', submodel, promotes=['*'])

        p.setup()
        p.final_setup()

        self.assertEqual(p.get_val('new_r'), 20)
        self.assertEqual(p.get_val('new_theta'), 0.5)
        self.assertEqual(p.get_val('new_x'), 100)

    def test_subprob_solver_print(self):
        p = om.Problem()
        subprob = om.Problem()
        subprob.model.add_subsystem('comp', om.ExecComp('x = r*cos(theta)'), promotes=['*'])
        submodel = om.SubmodelComp(problem=subprob)

        submodel.add_input('r', name='new_r', val=20)
        submodel.add_input('theta', name='new_theta', val=0.5)
        submodel.add_output('x', name='new_x', val=100)

        submodel._subprob.set_solver_print(level=3, depth=20, type_='NL')

        p.model.add_subsystem('submodel', submodel, promotes=['*'])

        self.assertTrue((3, 20, 'NL') in p.model.submodel._subprob.model._solver_print_cache)

    def test_complex_step_across_submodel(self):
        p = om.Problem()
        subprob = om.Problem()
        subprob.model.add_subsystem('comp', om.ExecComp('x = r*cos(theta)'), promotes=['*'])
        submodel = om.SubmodelComp(problem=subprob)

        submodel.add_input('r', name='new_r', val=20)
        submodel.add_input('theta', name='new_theta', val=0.5)
        submodel.add_output('x', name='new_x', val=100)

        model = p.model
        model.add_subsystem('submodel', submodel, promotes=['*'])
        model.add_design_var('new_r')
        model.add_design_var('new_theta')
        model.add_objective('new_x')

        p.setup(force_alloc_complex=True)
        p.run_model()

        totals = p.check_totals(method='cs')
        assert_check_totals(totals, atol=1e-11, rtol=1e-11)


@unittest.skipUnless(MPI and PETScVector, "MPI and PETSc are required.")
class TestSubmodelCompMPI(unittest.TestCase):
    N_PROCS = 2

    def test_submodel_comp(self):
        p = om.Problem()

        model = p.model

        par = model.add_subsystem('par', om.ParallelGroup())

        par.add_subsystem('subprob1', om.SubmodelComp(problem=om.Problem(model=FanOut()),
                                                      inputs=['p.x'], outputs=['comp2.y', 'comp3.y']))
        par.add_subsystem('subprob2', om.SubmodelComp(problem=om.Problem(model=FanIn()),
                                                      inputs=['p1.x1', 'p2.x2'], outputs=['comp3.y']))

        p.setup(force_alloc_complex=True)
        p.run_model()
        cpd = p.check_partials(method='cs', out_stream=None)
        assert_check_partials(cpd)


class IncompleteRelevanceGroup(om.Group):
    def __init__(self, size, **kwargs):
        super().__init__(**kwargs)
        self.size = size

    def setup(self):
        size = self.size
        self.add_subsystem('C1', om.ExecComp('y = 1.5*x', shape=size))
        self.add_subsystem('C2', om.ExecComp('y = 3.0*x', shape=size*2+1))
        self.add_subsystem('C3', om.ExecComp('y = 2.75*x', shape=size))
        self.add_subsystem('C4', om.ExecComp('y = 4.25*x1 - 0.75*x2', shape=size))
        self.add_subsystem('C5', om.ExecComp('y = sum(x)/sum(x**2)', x=np.ones(size*2+1), y=0.0))
        self.connect('C1.y', ['C3.x', 'C4.x1'])
        self.connect('C2.y', 'C4.x2', src_indices=list(range(size)))
        self.connect('C2.y', 'C5.x')


class TestSubmodelColoring(unittest.TestCase):
    def test_submodel_inner_coloring(self):
        # this one has coloring within the submodelcomp

        p = om.Problem()
        model = p.model


        model.add_subsystem('sub', om.SubmodelComp(problem=om.Problem(model=IncompleteRelevanceGroup(3)),
                                                    inputs=['C1.x', 'C2.x'], outputs=['C3.y', 'C4.y', 'C5.y']))

        p.setup(force_alloc_complex=True)
        p.run_model()

        check = p.check_partials(method='cs', show_only_incorrect=True)
        assert_check_partials(check)

        check = p.check_totals(of=['sub.C3:y', 'sub.C4:y', 'sub.C5:y'],
                               wrt=['sub.C1:x', 'sub.C2:x'], show_only_incorrect=True)
        assert_check_totals(check)

    def test_submodel_inner_outer_coloring(self):
        # this one has coloring within the submodelcomp and the outer problem driver

        p = om.Problem()
        model = p.model


        model.add_subsystem('sub', om.SubmodelComp(problem=om.Problem(model=IncompleteRelevanceGroup(3)),
                                                    inputs=['C1.x', 'C2.x'], outputs=['C3.y', 'C4.y', 'C5.y']))

        p.driver = om.ScipyOptimizeDriver(optimizer='SLSQP')
        p.driver.declare_coloring(show_summary=True)

        p.model.add_design_var('sub.C1:x')
        p.model.add_design_var('sub.C2:x')

        p.model.add_objective('sub.C5:y', index=0)
        p.model.add_constraint('sub.C4:y', lower=-1)
        p.model.add_constraint('sub.C3:y', upper=99)

        p.setup(force_alloc_complex=True)
        p.run_driver()

        check = p.check_partials(method='cs', show_only_incorrect=True)
        assert_check_partials(check)

        check = p.check_totals(of=['sub.C3:y', 'sub.C4:y', 'sub.C5:y'],
                               wrt=['sub.C1:x', 'sub.C2:x'], show_only_incorrect=True)
        assert_check_totals(check)

class TestSubmodelColoringMultiSubmodelComps(unittest.TestCase):

    def setUp(self):
        from openmdao.utils.general_utils import set_pyoptsparse_opt

        OPT, _ = set_pyoptsparse_opt('SLSQP')
        if OPT is None:
            raise unittest.SkipTest("pyoptsparse is not installed")

    def test_multiple_submodels_inner_outer_coloring(self):
        # this one has coloring within the submodelcomp and the outer problem driver

        p = om.Problem()

        model = p.model

        par = model.add_subsystem('par', om.ParallelGroup(), promotes=['*'])

        par.add_subsystem('sub1', om.SubmodelComp(problem=om.Problem(model=IncompleteRelevanceGroup(3)),
                                                  inputs=['C1.x', 'C2.x'], outputs=['C3.y', 'C4.y', 'C5.y']))
        par.add_subsystem('sub2', om.SubmodelComp(problem=om.Problem(model=IncompleteRelevanceGroup(3)),
                                                  inputs=['C1.x', 'C2.x'], outputs=['C3.y', 'C4.y', 'C5.y']))

        p.driver = om.ScipyOptimizeDriver(optimizer='SLSQP')
        p.driver.declare_coloring(show_summary=True)

        p.model.add_design_var('sub1.C1:x')
        p.model.add_design_var('sub1.C2:x')
        p.model.add_design_var('sub2.C1:x')
        p.model.add_design_var('sub2.C2:x')

        p.model.add_objective('sub1.C5:y', index=0)
        p.model.add_constraint('sub2.C5:y', lower=-1)
        p.model.add_constraint('sub1.C4:y', lower=-1)
        p.model.add_constraint('sub1.C3:y', upper=99)
        p.model.add_constraint('sub2.C4:y', lower=-1)
        p.model.add_constraint('sub2.C3:y', upper=99)

        p.setup(force_alloc_complex=True)
        p.run_driver()

        check = p.check_partials(method='cs', show_only_incorrect=True)
        assert_check_partials(check)

        check = p.check_totals(of=['sub1.C3:y', 'sub1.C4:y', 'sub1.C5:y',
                                      'sub2.C3:y', 'sub2.C4:y', 'sub2.C5:y'],
                                 wrt=['sub1.C1:x', 'sub1.C2:x', 'sub2.C1:x', 'sub2.C2:x'], show_only_incorrect=True)
        assert_check_totals(check)


@unittest.skipUnless(MPI and PETScVector, "MPI and PETSc are required.")
class TestSubmodelColoringMultiSubmodelCompsMPI(TestSubmodelColoringMultiSubmodelComps):
    N_PROCS = 2


class TestSubmodelOpt(unittest.TestCase):
    def test_submodel_with_opt(self):
        sub_prob = om.Problem()

        # modify paraboloid eqn with parameter "a" that will be constant during internal opt
        sub_prob.model.add_subsystem("paraboloid", om.ExecComp("f = (x-3)**2 + x*y + (y+4)**2 + a"))

        sub_prob.driver = om.ScipyOptimizeDriver(optimizer="SLSQP")

        sub_prob.model.add_design_var("paraboloid.x", lower=-50, upper=50)
        sub_prob.model.add_design_var("paraboloid.y", lower=-50, upper=50)
        sub_prob.model.add_objective("paraboloid.f")

        top_prob = om.Problem()
        submodel = om.SubmodelComp(problem=sub_prob, inputs=[("paraboloid.a", "a")], outputs=[("paraboloid.f", "g")],
                                   do_coloring=False)
        top_prob.model.add_subsystem("submodel", submodel, promotes=["*"])
        top_prob.setup()
        top_prob.set_val("a", val=-3)

        top_prob.run_model()

        assert_near_equal(sub_prob.get_val("paraboloid.f"), -27.33333307, tolerance=1e-8)
        assert_near_equal(sub_prob.get_val("paraboloid.x"), 6.66712855, tolerance=1e-8)
        assert_near_equal(sub_prob.get_val("paraboloid.y"), -7.33324946, tolerance=1e-8)

        # now test that we're not allowed to compute derivatives of a SubmodelComp if it has an optimizer
        with self.assertRaises(Exception) as cm:
            top_prob.compute_totals(of=['g'], wrt=['a'])

        self.assertEqual(cm.exception.args[0],
                         "'submodel' <class SubmodelComp>: Error calling compute_partials(), Can't compute partial "
                         "derivatives of a SubmodelComp with an internal optimizer.")


if __name__ == '__main__':
    unittest.main()



"""
This is a multipoint implementation of the beam optimization problem.

This version minimizes volume while satisfying a max bending stress constraint in each element
for each loadcase.
"""
import numpy as np

import openmdao.api as om

from openmdao.test_suite.test_examples.beam_optimization.components.local_stiffness_matrix_comp import LocalStiffnessMatrixComp
from openmdao.test_suite.test_examples.beam_optimization.components.moment_comp import MomentOfInertiaComp
from openmdao.test_suite.test_examples.beam_optimization.components.multi_states_comp import MultiStatesComp
from openmdao.test_suite.test_examples.beam_optimization.components.multi_stress_comp import MultiStressComp
from openmdao.test_suite.test_examples.beam_optimization.components.volume_comp import VolumeComp
from openmdao.utils.spline_distributions import sine_distribution


def divide_cases(ncases, nprocs):
    """
    Divide up load cases among available procs.

    Parameters
    ----------
    ncases : int
        Number of load cases.
    nprocs : int
        Number of processors.

    Returns
    -------
    list of list of int
        Integer case numbers for each proc.
    """
    data = []
    for j in range(nprocs):
        data.append([])

    wrap = 0
    for j in range(ncases):
        idx = j - wrap
        if idx >= nprocs:
            idx = 0
            wrap = j

        data[idx].append(j)

    return data


class MultipointBeamGroup(om.Group):
    """
    System setup for minimization of volume (i.e., mass) subject to KS aggregated bending stress constraints.
    """

    def initialize(self):
        self.options.declare('E')
        self.options.declare('L')
        self.options.declare('b')
        self.options.declare('volume')
        self.options.declare('max_bending')
        self.options.declare('num_elements', 5)
        self.options.declare('num_cp', 50)
        self.options.declare('num_load_cases', 1)
        self.options.declare('parallel_derivs', False, types=bool, allow_none=True)
        self.options.declare('ks_add_constraint', default=False, types=bool)

    def setup(self):
        E = self.options['E']
        L = self.options['L']
        b = self.options['b']
        volume = self.options['volume']
        max_bending = self.options['max_bending']
        num_elements = self.options['num_elements']
        num_nodes = num_elements + 1
        num_cp = self.options['num_cp']
        num_load_cases = self.options['num_load_cases']
        parallel_derivs = self.options['parallel_derivs']

        x_interp = sine_distribution(num_elements)
        comp = om.SplineComp(method='bsplines', num_cp=num_cp, x_interp_val=x_interp)
        comp.add_spline(y_cp_name='h_cp', y_interp_name='h')
        self.add_subsystem('interp', comp)

        I_comp = MomentOfInertiaComp(num_elements=num_elements, b=b)
        self.add_subsystem('I_comp', I_comp)

        comp = LocalStiffnessMatrixComp(num_elements=num_elements, E=E, L=L)
        self.add_subsystem('local_stiffness_matrix_comp', comp)

        # Parallel Subsystem for load cases.
        par = self.add_subsystem('parallel', om.ParallelGroup())

        # Determine how to split cases up over the available procs.
        nprocs = self.comm.size
        divide = divide_cases(num_load_cases, nprocs)

        for j, this_proc in enumerate(divide):
            num_rhs = len(this_proc)

            name = 'sub_%d' % j
            sub = par.add_subsystem(name, om.Group())

            # Load is a sinusoidal distributed force of varying spatial frequency.
            force_vector = np.zeros((2 * num_nodes, num_rhs))
            for i, k in enumerate(this_proc):

                end = 1.5 * np.pi
                if num_load_cases > 1:
                    end += k * 0.5 * np.pi / (num_load_cases - 1)

                x = np.linspace(0, end, num_nodes)
                f = - np.sin(x)
                force_vector[0:-1:2, i] = f

            comp = MultiStatesComp(num_elements=num_elements, force_vector=force_vector,
                                   num_rhs=num_rhs)
            sub.add_subsystem('states_comp', comp)

            comp = MultiStressComp(num_elements=num_elements, E=E, num_rhs=num_rhs)
            sub.add_subsystem('stress_comp', comp)

            self.connect('local_stiffness_matrix_comp.K_local',
                         'parallel.%s.states_comp.K_local' % name)

            for k in range(num_rhs):
                sub.connect('states_comp.d_%d' % k,
                            'stress_comp.displacements_%d' % k,
                            src_indices=np.arange(2 *num_nodes))

                if parallel_derivs:
                    color = 'red_%d' % k
                else:
                    color = None

                comp = om.KSComp(width=num_elements, upper=max_bending,
                                 add_constraint=self.options['ks_add_constraint'],
                                 parallel_deriv_color=color)

                sub.add_subsystem('KS_%d' % k, comp)

                sub.connect('stress_comp.stress_%d' % k,
                            'KS_%d.g' % k)

                if not self.options['ks_add_constraint']:
                    sub.add_constraint('KS_%d.KS' % k, upper=0.0,
                                       parallel_deriv_color=color)

        comp = VolumeComp(num_elements=num_elements, b=b, L=L)
        self.add_subsystem('volume_comp', comp)

        self.connect('interp.h', 'I_comp.h')
        self.connect('interp.h', 'volume_comp.h')
        self.connect('I_comp.I', 'local_stiffness_matrix_comp.I')

        self.add_design_var('interp.h_cp', lower=1e-2, upper=10.)
        self.add_objective('volume_comp.volume')



"""Define the PETSc Transfer class."""
import numpy as np
from openmdao.utils.mpi import check_mpi_env
from openmdao.core.constants import INT_DTYPE

use_mpi = check_mpi_env()
_empty_idx_array = np.array([], dtype=INT_DTYPE)


if use_mpi is False:
    PETScTransfer = None
else:
    try:
        import petsc4py
        from petsc4py import PETSc
    except ImportError:
        PETSc = None
        if use_mpi is True:
            raise ImportError("Importing petsc4py failed and OPENMDAO_USE_MPI is true.")

    from petsc4py import PETSc
    from collections import defaultdict

    from openmdao.vectors.default_transfer import DefaultTransfer, _setup_index_views

    class PETScTransfer(DefaultTransfer):
        """
        PETSc Transfer implementation for running in parallel.

        Parameters
        ----------
        in_vec : <Vector>
            Pointer to the input vector.
        out_vec : <Vector>
            Pointer to the output vector.
        in_inds : int ndarray
            Input indices for the transfer.
        out_inds : int ndarray
            Output indices for the transfer.
        comm : MPI.Comm or <FakeComm>
            Communicator of the system that owns this transfer.

        Attributes
        ----------
        _scatter : method
            Method that performs a PETSc scatter.
        """

        def __init__(self, in_vec, out_vec, in_inds, out_inds, comm):
            """
            Initialize all attributes.
            """
            super().__init__(in_vec, out_vec, in_inds, out_inds)
            in_indexset = PETSc.IS().createGeneral(self._in_inds, comm=comm)
            out_indexset = PETSc.IS().createGeneral(self._out_inds, comm=comm)

            self._scatter = PETSc.Scatter().create(out_vec._petsc, out_indexset, in_vec._petsc,
                                                   in_indexset).scatter

        @staticmethod
        def _setup_transfers(group):
            """
            Compute all transfers that are owned by our parent group.

            Parameters
            ----------
            group : <Group>
                Parent group.
            """
            rev = group._orig_mode != 'fwd'

            group._transfers = {
                'fwd': PETScTransfer._setup_transfers_fwd(group)
            }

            if rev:
                group._transfers['rev'] = PETScTransfer._setup_transfers_rev(group)

        @staticmethod
        def _setup_transfers_fwd(group):
            transfers = {}

            if not group._conn_abs_in2out:
                return transfers

            abs2meta_in = group._var_abs2meta['input']
            myrank = group.comm.rank

            offsets_in = group._get_var_offsets()['input'][myrank, :]
            mypathlen = len(group.pathname) + 1 if group.pathname else 0

            xfer_in = defaultdict(list)
            xfer_out = defaultdict(list)

            allprocs_abs2idx = group._var_allprocs_abs2idx
            sizes_in = group._var_sizes['input'][myrank, :]

            total_len = 0

            # Loop through all connections owned by this system
            for abs_in, abs_out in group._conn_abs_in2out.items():
                sub_in = abs_in[mypathlen:].partition('.')[0]

                # Only continue if the input exists on this processor
                if abs_in in abs2meta_in:

                    output_inds, _ = _get_output_inds(group, abs_out, abs_in)

                    idx_in = allprocs_abs2idx[abs_in]
                    input_inds = range(offsets_in[idx_in], offsets_in[idx_in] + sizes_in[idx_in])

                    total_len += len(input_inds)

                    xfer_in[sub_in].append(input_inds)
                    xfer_out[sub_in].append(output_inds)
                else:
                    # not a local input but still need entries in the transfer dicts to
                    # avoid hangs
                    xfer_in[sub_in]  # defaultdict will create an empty list there
                    xfer_out[sub_in]

            if xfer_in:
                full_xfer_in, full_xfer_out = _setup_index_views(total_len, xfer_in, xfer_out)
                # full transfer (transfer to all subsystems at once)
                transfers[None] = PETScTransfer(group._vectors['input']['nonlinear'],
                                                group._vectors['output']['nonlinear'],
                                                full_xfer_in, full_xfer_out, group.comm)

                # transfers to individual subsystems
                for sname, inds in xfer_in.items():
                    transfers[sname] = PETScTransfer(group._vectors['input']['nonlinear'],
                                                     group._vectors['output']['nonlinear'],
                                                     inds, xfer_out[sname], group.comm)

            return transfers

        @staticmethod
        def _setup_transfers_rev(group):
            abs2meta_in = group._var_abs2meta['input']
            abs2meta_out = group._var_abs2meta['output']
            allprocs_abs2prom = group._var_allprocs_abs2prom

            # for an FD group, we use the relevance graph to determine which inputs on the
            # boundary of the group are upstream of responses within the group so
            # that we can perform any necessary corrections to the derivative inputs.
            if group._owns_approx_jac:
                if group.comm.size > 1 and group.pathname != '' and group._has_distrib_vars:
                    all_abs2meta_out = group._var_allprocs_abs2meta['output']
                    all_abs2meta_in = group._var_allprocs_abs2meta['input']

                    # connections internal to this group and all direct/indirect subsystems
                    conns = group._conn_global_abs_in2out

                    inp_boundary_set = set(all_abs2meta_in).difference(conns)

                    if inp_boundary_set:
                        for dv, resp, rel in group._relevance.iter_seed_pair_relevance(inputs=True):
                            if resp in all_abs2meta_out and dv not in allprocs_abs2prom:
                                # response is continuous and inside this group and
                                # dv is outside this group
                                if all_abs2meta_out[resp]['distributed']:  # a distributed response
                                    for inp in inp_boundary_set.intersection(rel):
                                        if inp in abs2meta_in:
                                            if resp not in group._fd_rev_xfer_correction_dist:
                                                group._fd_rev_xfer_correction_dist[resp] = set()
                                            group._fd_rev_xfer_correction_dist[resp].add(inp)

                # FD groups don't need reverse transfers
                return {}

            myrank = group.comm.rank
            allprocs_abs2idx = group._var_allprocs_abs2idx
            transfers = group._transfers
            vectors = group._vectors
            offsets = group._get_var_offsets()
            mypathlen = len(group.pathname) + 1 if group.pathname else 0

            has_par_coloring = group._problem_meta['has_par_deriv_color']

            xfer_in = defaultdict(list)
            xfer_out = defaultdict(list)

            # xfers that are only active when parallel coloring is not
            xfer_in_nocolor = defaultdict(list)
            xfer_out_nocolor = defaultdict(list)

            sizes_in = group._var_sizes['input']
            offsets_in = offsets['input']
            offsets_out = offsets['output']

            total_size = total_size_nocolor = 0

            # Loop through all connections owned by this system
            for abs_in, abs_out in group._conn_abs_in2out.items():
                sub_out = abs_out[mypathlen:].partition('.')[0]

                # Only continue if the input exists on this processor
                if abs_in in abs2meta_in:
                    meta_in = abs2meta_in[abs_in]
                    idx_in = allprocs_abs2idx[abs_in]
                    idx_out = allprocs_abs2idx[abs_out]

                    output_inds, src_indices = _get_output_inds(group, abs_out, abs_in)

                    # 2. Compute the input indices
                    input_inds = range(offsets_in[myrank, idx_in],
                                       offsets_in[myrank, idx_in] + sizes_in[myrank, idx_in])

                    # Now the indices are ready - input_inds, output_inds
                    inp_is_dup, inp_missing, distrib_in = group.get_var_dup_info(abs_in, 'input')
                    out_is_dup, _, distrib_out = group.get_var_dup_info(abs_out, 'output')

                    iowninput = myrank == group._owning_rank[abs_in]

                    if inp_is_dup and (abs_out not in abs2meta_out or
                                       (distrib_out and not iowninput)):
                        xfer_in[sub_out]
                        xfer_out[sub_out]
                    elif out_is_dup and inp_missing > 0 and (iowninput or distrib_in):
                        # if this rank owns the input or the input is distributed,
                        # and the output is duplicated, then we send the owning/distrib input
                        # to each duplicated output that doesn't have a corresponding connected
                        # input on the same rank.
                        oidxlist = []
                        iidxlist = []
                        oidxlist_nc = []
                        iidxlist_nc = []
                        size = size_nc = 0
                        for rnk, osize, isize in zip(range(group.comm.size),
                                                     group.get_var_sizes(abs_out, 'output'),
                                                     group.get_var_sizes(abs_in, 'input')):
                            if rnk == myrank:  # transfer to output on same rank
                                oidxlist.append(output_inds)
                                iidxlist.append(input_inds)
                                size += len(input_inds)
                            elif osize > 0 and isize == 0:
                                # dup output exists on this rank but there is no corresponding
                                # input, so we send the owning/distrib input to the dup output
                                offset = offsets_out[rnk, idx_out]
                                if src_indices is None:
                                    oarr = range(offset, offset + meta_in['size'])
                                elif src_indices.size > 0:
                                    oarr = np.asarray(src_indices + offset, dtype=INT_DTYPE)
                                else:
                                    continue

                                if has_par_coloring:
                                    # these transfers will only happen if parallel coloring is
                                    # not active for the current seed response
                                    oidxlist_nc.append(oarr)
                                    iidxlist_nc.append(input_inds)
                                    size_nc += len(input_inds)
                                else:
                                    oidxlist.append(oarr)
                                    iidxlist.append(input_inds)
                                    size += len(input_inds)

                        if len(iidxlist) > 1:
                            input_inds = _merge(iidxlist, size)
                            output_inds = _merge(oidxlist, size)
                        else:
                            input_inds = iidxlist[0]
                            output_inds = oidxlist[0]

                        total_size += len(input_inds)

                        xfer_in[sub_out].append(input_inds)
                        xfer_out[sub_out].append(output_inds)

                        if has_par_coloring and iidxlist_nc:
                            # keep transfers separate that shouldn't happen when parallel
                            # deriv coloring is active
                            if len(iidxlist_nc) > 1:
                                input_inds = _merge(iidxlist_nc, size_nc)
                                output_inds = _merge(oidxlist_nc, size_nc)
                            else:
                                input_inds = iidxlist_nc[0]
                                output_inds = oidxlist_nc[0]

                            total_size_nocolor += len(input_inds)

                            xfer_in_nocolor[sub_out].append(input_inds)
                            xfer_out_nocolor[sub_out].append(output_inds)
                    else:
                        if (inp_is_dup and out_is_dup and src_indices is not None and
                                src_indices.size > 0):
                            offset = offsets_out[myrank, idx_out]
                            output_inds = np.asarray(src_indices + offset, dtype=INT_DTYPE)

                        total_size += len(input_inds)

                        xfer_in[sub_out].append(input_inds)
                        xfer_out[sub_out].append(output_inds)
                else:
                    # remote input but still need entries in the transfer dicts to avoid hangs
                    xfer_in[sub_out]
                    xfer_out[sub_out]
                    if has_par_coloring:
                        xfer_in_nocolor[sub_out]
                        xfer_out_nocolor[sub_out]

            full_xfer_in, full_xfer_out = _setup_index_views(total_size, xfer_in, xfer_out)

            transfers = {
                None: PETScTransfer(vectors['input']['nonlinear'],
                                    vectors['output']['nonlinear'],
                                    full_xfer_in, full_xfer_out, group.comm)
            }

            for sname, inds in xfer_out.items():
                transfers[sname] = PETScTransfer(vectors['input']['nonlinear'],
                                                 vectors['output']['nonlinear'],
                                                 xfer_in[sname], inds, group.comm)

            if xfer_in_nocolor:
                full_xfer_in, full_xfer_out = _setup_index_views(total_size_nocolor,
                                                                 xfer_in_nocolor,
                                                                 xfer_out_nocolor)

                transfers[(None, '@nocolor')] = PETScTransfer(vectors['input']['nonlinear'],
                                                              vectors['output']['nonlinear'],
                                                              full_xfer_in, full_xfer_out,
                                                              group.comm)

                for sname, inds in xfer_out_nocolor.items():
                    transfers[(sname, '@nocolor')] = PETScTransfer(vectors['input']['nonlinear'],
                                                                   vectors['output']['nonlinear'],
                                                                   xfer_in_nocolor[sname], inds,
                                                                   group.comm)

            return transfers

        def _transfer(self, in_vec, out_vec, mode='fwd'):
            """
            Perform transfer.

            Parameters
            ----------
            in_vec : <Vector>
                pointer to the input vector.
            out_vec : <Vector>
                pointer to the output vector.
            mode : str
                'fwd' or 'rev'.
            """
            flag = False
            if mode == 'rev':
                flag = True
                in_vec, out_vec = out_vec, in_vec

            in_petsc = in_vec._petsc
            out_petsc = out_vec._petsc

            # For Complex Step, need to disassemble real and imag parts, transfer them separately,
            # then reassemble them.
            if in_vec._under_complex_step and out_vec._alloc_complex:

                # Real
                in_petsc.array = in_vec._data.real
                out_petsc.array = out_vec._data.real
                self._scatter(out_petsc, in_petsc, addv=flag, mode=flag)

                # Imaginary
                in_petsc_imag = in_vec._imag_petsc
                out_petsc_imag = out_vec._imag_petsc
                in_petsc_imag.array = in_vec._data.imag
                out_petsc_imag.array = out_vec._data.imag
                self._scatter(out_petsc_imag, in_petsc_imag, addv=flag, mode=flag)

                in_vec._data[:] = in_petsc.array + in_petsc_imag.array * 1j

            else:

                # Anything that has been allocated complex requires an additional step because
                # the petsc vector does not directly reference the _data.

                if in_vec._alloc_complex:
                    in_petsc.array = in_vec._get_data()

                if out_vec._alloc_complex:
                    out_petsc.array = out_vec._get_data()

                self._scatter(out_petsc, in_petsc, addv=flag, mode=flag)

                if in_vec._alloc_complex:
                    data = in_vec._get_data()
                    data[:] = in_petsc.array


def _merge(inds_list, tot_size):
    """
    Convert a list of indices and/or ranges into an array.

    Parameters
    ----------
    inds_list : list of ranges or ndarrays
        List of indices.
    tot_size : int
        Total size of the indices in the list.

    Returns
    -------
    ndarray
        Array of indices.
    """
    if inds_list:
        arr = np.empty(tot_size, dtype=INT_DTYPE)
        start = end = 0
        for inds in inds_list:
            end += len(inds)
            arr[start:end] = inds
            start = end

        return arr

    return _empty_idx_array


def _get_output_inds(group, abs_out, abs_in):
    owner = group._owning_rank[abs_out]
    meta_in = group._var_abs2meta['input'][abs_in]
    out_dist = group._var_allprocs_abs2meta['output'][abs_out]['distributed']
    in_dist = meta_in['distributed']
    src_indices = meta_in['src_indices']

    rank = group.comm.rank if abs_out in group._var_abs2meta['output'] else owner
    out_idx = group._var_allprocs_abs2idx[abs_out]
    offsets = group._get_var_offsets()['output'][:, out_idx]
    sizes = group._var_sizes['output'][:, out_idx]

    if src_indices is None:
        orig_src_inds = src_indices
    else:
        src_indices = src_indices.shaped_array()
        orig_src_inds = src_indices
        if not out_dist and not in_dist:  # convert from local to distributed src_indices
            off = np.sum(sizes[:rank])
            if off > 0.:  # adjust for local offsets
                # don't do += to avoid modifying stored value
                src_indices = src_indices + off

    # NOTE: src_indices are relative to a single, possibly distributed variable,
    # while the output_inds that we compute are relative to the full distributed
    # array that contains all local variables from each rank stacked in rank order.
    if src_indices is None:
        if out_dist:
            # input in this case is non-distributed (else src_indices would be
            # defined by now).  dist output to non-distributed input conns w/o
            # src_indices are not allowed.
            raise RuntimeError(f"{group.msginfo}: Can't connect distributed output "
                               f"'{abs_out}' to non-distributed input '{abs_in}' "
                               "without declaring src_indices.", ident=(abs_out, abs_in))
        else:
            offset = offsets[rank]
            output_inds = range(offset, offset + sizes[rank])
    else:
        output_inds = np.empty(src_indices.size, INT_DTYPE)
        start = end = 0
        for iproc in range(group.comm.size):
            end += sizes[iproc]
            if start == end:
                continue

            # The part of src on iproc
            on_iproc = np.logical_and(start <= src_indices, src_indices < end)

            if np.any(on_iproc):
                # This converts from global to variable specific ordering
                output_inds[on_iproc] = src_indices[on_iproc] + (offsets[iproc] - start)

            start = end

    return output_inds, orig_src_inds



"""
Various graph related utilities.
"""
import networkx as nx


def get_sccs_topo(graph):
    """
    Return strongly connected subsystems of the given Group in topological order.

    Parameters
    ----------
    graph : networkx.DiGraph
        Directed graph of Systems.

    Returns
    -------
    list of sets of str
        A list of strongly connected components in topological order.
    """
    # Tarjan's algorithm returns SCCs in reverse topological order, so
    # the list returned here is reversed.
    sccs = list(nx.strongly_connected_components(graph))
    sccs.reverse()
    return sccs


def get_out_of_order_nodes(graph, orders):
    """
    Return a list of nodes that are out of order.

    Parameters
    ----------
    graph : networkx.DiGraph
        Directed graph of Systems.
    orders : dict
        A dict of order values keyed by node name.

    Returns
    -------
    list of sets of str
        A list of strongly connected components in topological order.
    list of str
        A list of nodes that are out of order.
    """
    strongcomps = get_sccs_topo(graph)

    out_of_order = []
    for strongcomp in strongcomps:
        for u, v in graph.edges(strongcomp):
            # for any connection between a system in this strongcomp and a system
            # outside of it, the target must be ordered after the source.
            if u in strongcomp and v not in strongcomp and orders[u] > orders[v]:
                out_of_order.append((u, v))

    return strongcomps, out_of_order
